import os

import pandas as pd

os.environ["KMP_DUPLICATE_LIB_OK"]="TRUE"
from typing import List, Optional, Tuple, Union

import numpy as np

from simba.data_processors.cuda.utils import _is_cuda_available

try:
    from ultralytics.models.sam import SAM2VideoPredictor
except:
    SAM2VideoPredictor = None

from simba.utils.checks import (check_file_exist_and_readable, check_float,
                                check_if_dir_exists, check_instance, check_int,
                                check_valid_array, check_valid_tuple)
from simba.utils.data import resample_geometry_vertices
from simba.utils.enums import Formats, Options
from simba.utils.errors import (InvalidInputError, SimBAGPUError,
                                SimBAPAckageVersionError)
from simba.utils.read_write import (find_files_of_filetypes_in_directory,
                                    get_fn_ext, get_video_meta_data, write_df)


class SamInference():

    """
    :example:
    >>> i = SamInference(video_path=r"MyVideo",
    >>>                 labels=[[1]],
    >>>                 prompts=[[166, 428]],
    >>>                 weights_path=r"D:\yolo_weights\sam2.1_b.pt",
    >>>                 save_dir=r'C:\troubleshooting\sam_results',
    >>>                 names=('Animal1',))
    >>> i.run()


    .. video:: _static/img/sam_example.webm
       :loop:

    """
    def __init__(self,
                 video_path: Union[str, os.PathLike],
                 weights_path: Union[str, os.PathLike],
                 save_dir: Union[str, os.PathLike],
                 prompts: Union[np.ndarray, List[List[int]]],
                 labels: Union[np.ndarray, List[List[int]]],
                 names: Tuple[str, ...],
                 imgsz: Optional[int] = 1024,
                 confidence: Optional[float] = 0.25,
                 vertice_cnt: Optional[int] = 100):

        if not _is_cuda_available()[0]:
            raise SimBAGPUError(msg='No GPU detected.', source=self.__class__.__name__)
        if SAM2VideoPredictor is None:
            raise SimBAPAckageVersionError(msg='ultralytics.models.sam.SAM2VideoPredictor package not detected.', source=self.__class__.__name__)
        if os.path.isdir(video_path):
            self.video_paths = find_files_of_filetypes_in_directory(directory=video_path, extensions=Options.ALL_VIDEO_FORMAT_OPTIONS.value, raise_error=True)
        elif os.path.isfile(video_path):
            self.video_paths = [video_path]
        else:
            raise InvalidInputError(msg=f'{video_path} is not a valid file or directory path.', source=self.__class__.__name__)
        _ = [get_video_meta_data(video_path=x) for x in self.video_paths]
        check_instance(source=f'{self.__class__.__name__} prompts', instance=prompts, accepted_types=(list, np.ndarray,), raise_error=True, warning=False)
        check_instance(source=f'{self.__class__.__name__} labels', instance=labels, accepted_types=(list, np.ndarray,), raise_error=True, warning=False)
        prompts = np.array(prompts) if isinstance(prompts, (list,)) else prompts
        labels = np.array(labels) if isinstance(labels, (list,)) else labels
        check_valid_array(data=prompts, source=f'{self.__class__.__name__} prompts', accepted_ndims=(1, 2, 3,), accepted_dtypes=Formats.INTEGER_DTYPES.value)
        check_valid_array(data=labels, source=f'{self.__class__.__name__} labels', accepted_ndims=(1, 2,), accepted_dtypes=Formats.INTEGER_DTYPES.value)
        check_file_exist_and_readable(file_path=weights_path)
        check_if_dir_exists(in_dir=save_dir, raise_error=True)
        check_int(name=f'{self.__class__.__name__} imgsz', value=imgsz, min_value=1)
        check_int(name=f'{self.__class__.__name__} vertice_cnt', value=vertice_cnt, min_value=3)
        check_float(name=f'{self.__class__.__name__} confidence', value=confidence, min_value=10e-6, max_value=1.0)
        check_valid_tuple(x=names, source=f'{self.__class__.__name__} names', accepted_lengths=(len(labels[0]),))
        self.animal_name_dict = {v: x for v, x in enumerate(names)}
        self.prompts, self.lbls, self.vertice_cnt = prompts, labels, vertice_cnt
        self.save_dir, self.names = save_dir, names
        self.overrides = dict(conf=confidence, task="segment", mode="predict", imgsz=imgsz, model="sam2_b.pt")
        self.predictor = SAM2VideoPredictor(overrides=self.overrides)
        self.vertice_col_names = ['FRAME', 'NAME']
        for i in range(self.vertice_cnt):
            self.vertice_col_names.append(f"VERTICE_{i}_x"); self.vertice_col_names.append(f"VERTICE_{i}_y")

    def run(self):
        for video_cnt, video_path in enumerate(self.video_paths):
            video_labels, video_prompts = self.lbls[video_cnt], self.prompts[video_cnt]
            video_meta_data = get_video_meta_data(video_path=video_path)
            video_results = []
            _, video_name, _ = get_fn_ext(filepath=video_path)
            save_path = os.path.join(self.save_dir, f'{video_name}.csv')
            results = self.predictor(source=video_path, points=video_prompts, labels=video_labels, stream=True)
            for frm_cnt, video_predictions in enumerate(results):
                for animal_name_cnt, animal_name in enumerate(self.names):
                    if video_predictions.names is None or animal_name_cnt not in video_predictions.names.keys():
                        mask = np.full(shape=(int(self.vertice_cnt*2)), fill_value=-1, dtype=np.int32)
                        mask = np.insert(mask, 0, animal_name_cnt)
                        mask = np.insert(mask, 0, int(frm_cnt))
                        video_results.append(mask)
                    else:
                        mask = video_predictions.masks[animal_name_cnt].xy[0].astype(np.int64)
                        mask[:, 0] = np.clip(mask[:, 0], 0, video_meta_data['width'])
                        mask[:, 1] = np.clip(mask[:, 1], 0, video_meta_data['height'])
                        mask = resample_geometry_vertices(vertices=mask.reshape(1, len(mask), 2), vertice_cnt=self.vertice_cnt)[0].flatten().astype(np.int64)
                        mask = np.insert(mask, 0, animal_name_cnt)
                        mask = np.insert(mask, 0, int(frm_cnt))
                        video_results.append(mask)
            video_results = pd.DataFrame(video_results, columns=self.vertice_col_names)
            #video_results['NAME'] = video_results['NAME'].map(self.animal_name_dict)
            video_results.to_csv(path_or_buf=save_path)
            #write_df(df=video_results, file_type='csv', save_path=save_path, multi_idx_header=False)






# i = SamInference(video_path=r"D:\platea\platea_videos\videos\clipped\10B_Mouse_5-choice_MustTouchTrainingNEWFINAL_a7_clipped_3.mp4",
#                  labels=[[1]],
#                  prompts=[[166, 428]],
#                  weights_path=r"D:\yolo_weights\sam2.1_b.pt",
#                  save_dir=r'C:\troubleshooting\sam_results',
#                  names=('Animal1',))
# i.run()


__author__ = "Simon Nilsson"

import os
from copy import deepcopy
from tkinter import *
from typing import Union

from simba.data_processors.smoothing import Smoothing
from simba.mixins.config_reader import ConfigReader
from simba.mixins.pop_up_mixin import PopUpMixin
from simba.ui.tkinter_functions import (DropDownMenu, Entry_Box, FileSelect,
                                        FolderSelect, SimbaButton)
from simba.utils.checks import (check_file_exist_and_readable,
                                check_if_dir_exists, check_int)
from simba.utils.enums import Formats, Options
from simba.utils.read_write import str_2_bool

SMOOTHING_OPTION = {'Savitzky Golay': "savitzky-golay", "Gaussian": "gaussian"}

class SmoothingPopUp(PopUpMixin, ConfigReader):

    """
    :example:
    >>> SmoothingPopUp(config_path='/Users/simon/Desktop/envs/simba/troubleshooting/two_black_animals_14bp/project_folder/project_config.ini')
    """
    def __init__(self, config_path: Union[str, os.PathLike]):
        PopUpMixin.__init__(self, title="SMOOTH POSE-ESTIMATION DATA", icon='smooth')
        ConfigReader.__init__(self, config_path=config_path, read_video_info=False)
        self.config_path = config_path

        self.settings_frm = LabelFrame(self.main_frm, text="SETTINGS", font=Formats.FONT_HEADER.value)
        self.time_window = Entry_Box(self.settings_frm, "TIME WINDOW (MILLISECONDS):", "35", validation="numeric")
        self.method_dropdown = DropDownMenu(self.settings_frm, "SMOOTHING METHOD:", Options.SMOOTHING_OPTIONS.value, "35")
        self.save_originals_dropdown = DropDownMenu(self.settings_frm, "SAVE ORIGINALS:", Options.BOOL_STR_OPTIONS.value, "35")
        self.save_originals_dropdown.setChoices(Options.BOOL_STR_OPTIONS.value[0])

        self.method_dropdown.setChoices(Options.SMOOTHING_OPTIONS.value[0])
        self.settings_frm.grid(row=0, column=0, sticky=NW)
        self.time_window.grid(row=0, column=0, sticky=NW)
        self.method_dropdown.grid(row=1, column=0, sticky=NW)
        self.save_originals_dropdown.grid(row=2, column=0, sticky=NW)

        self.single_file_frm = LabelFrame(self.main_frm, text="SMOOTH SINGLE DATA FILE", font=Formats.FONT_HEADER.value)
        self.selected_file = FileSelect(self.single_file_frm, "DATA PATH:", lblwidth=35, file_types=[("VIDEO FILE", ".csv .parquet")], initialdir=self.project_path)


        self.run_btn_single = SimbaButton(parent=self.single_file_frm, txt="RUN SINGLE DATA FILE SMOOTHING", img='rocket', txt_clr="blue", font=Formats.FONT_REGULAR.value, cmd=self.run, cmd_kwargs={'multiple': False})
        self.single_file_frm.grid(row=1, column=0, sticky=NW)
        self.selected_file.grid(row=0, column=0, sticky=NW)
        self.run_btn_single.grid(row=1, column=0, sticky=NW)

        self.multiple_file_frm = LabelFrame(self.main_frm, text="SMOOTH DIRECTORY OF DATA", font=Formats.FONT_HEADER.value)
        self.selected_dir = FolderSelect(self.multiple_file_frm, "SELECT DIRECTORY OF DATA FILES:", lblwidth=35, initialdir=self.project_path)

        self.run_btn_multiple = SimbaButton(parent=self.multiple_file_frm, txt="RUN DATA DIRECTORY SMOOTHING", img='rocket', txt_clr="blue", font=Formats.FONT_REGULAR.value, cmd=self.run, cmd_kwargs={'multiple': True})
        self.multiple_file_frm.grid(row=2, column=0, sticky=NW)
        self.selected_dir.grid(row=0, column=0, sticky=NW)
        self.run_btn_multiple.grid(row=1, column=0, sticky=NW)
        self.main_frm.mainloop()

    def run(self, multiple):
        smooth_time = self.time_window.entry_get
        smooth_method = SMOOTHING_OPTION[self.method_dropdown.getChoices()]
        copy_originals = str_2_bool(self.save_originals_dropdown.getChoices())
        check_int(name='TIME WINDOW (MILLISECONDS)', value=smooth_time, min_value=1)

        if not multiple:
            data_path = self.selected_file.file_path
            check_file_exist_and_readable(file_path=data_path)
            data_dir = os.path.dirname(data_path)
        else:
            data_path = self.selected_dir.folder_path
            check_if_dir_exists(in_dir=data_path)
            data_dir = deepcopy(data_path)
        multi_index_df_headers = False
        if data_dir == self.input_csv_dir: multi_index_df_headers = True

        smoothing = Smoothing(config_path=self.config_path,
                              data_path=data_path,
                              time_window=int(smooth_time),
                              method=smooth_method,
                              multi_index_df_headers=multi_index_df_headers,
                              copy_originals=copy_originals)
        smoothing.run()


#SmoothingPopUp(config_path='"C:\troubleshooting\mitra\project_folder\project_config.ini"')



import os
from typing import Optional, Union

import numpy as np
import pandas as pd
from numba import typed

from simba.mixins.config_reader import ConfigReader
from simba.mixins.feature_extraction_mixin import FeatureExtractionMixin
from simba.mixins.timeseries_features_mixin import TimeseriesFeatureMixin
from simba.utils.checks import (
    check_all_file_names_are_represented_in_video_log, check_if_dir_exists,
    check_int, check_str, check_valid_dataframe)
from simba.utils.data import detect_bouts, plug_holes_shortest_bout
from simba.utils.enums import Formats
from simba.utils.printing import stdout_success
from simba.utils.read_write import (find_files_of_filetypes_in_directory,
                                    get_fn_ext, read_df, read_video_info)

NAPE_X, NAPE_Y = 'nape_x', 'nape_y'
FREEZING = 'FREEZING'

class FreezingDetector(ConfigReader):

    """
    Detect freezing behavior using heuristic rules.

    .. important::

        Freezing is detected as `present` when **the velocity (computed from the mean movement of the nape, nose, and tail-base body-parts) falls below
        the movement threshold for the duration (and longer) of the specied time-window**.

        Freezing is detected as `absent` when not present.

    .. note::

       We pass the names of the left and right ears, as the method will use body-parts to compute the `nape` location of the animal.

    :param Union[str, os.PathLike] data_dir: Path to directory containing pose-estimated body-part data in CSV format.
    :param Union[str, os.PathLike] config_path: Path to SimBA project config file.
    :param Optional[str] nose_name: The name of the pose-estimated nose body-part. Defaults to 'nose'.
    :param Optional[str] left_ear_name: The name of the pose-estimated left ear body-part. Defaults to 'left_ear'.
    :param Optional[str] right_ear_name: The name of the pose-estimated right ear body-part. Defaults to 'right_ear'.
    :param Optional[str] tail_base_name: The name of the pose-estimated tail base body-part. Defaults to 'tail_base'.
    :param Optional[int] time_window: The time window in preceding seconds in which to evaluate freezing. Default: 3.
    :param Optional[int] movement_threshold: A movement threshold in millimeters per second. Defaults to 5.
    :param Optional[Union[str, os.PathLike]] save_dir: Directory where to store the results. If None, then results are stored in the ``logs`` directory of the SimBA project.

    :example:
    >>> FreezingDetector(data_dir=r'D:\troubleshooting\mitra\project_folder\csv\outlier_corrected_movement_location', config_path=r"D:\troubleshooting\mitra\project_folder\project_config.ini")

    References
    ----------
    .. [1] Sabnis et al., Visual detection of seizures in mice using supervised machine learning, `biorxiv`, doi: https://doi.org/10.1101/2024.05.29.596520.
    .. [2] Lopez et al., Region-specific Nucleus Accumbens Dopamine Signals Encode Distinct Aspects of Avoidance Learning, `biorxiv`, doi: https://doi.org/10.1101/2024.08.28.610149
    .. [3] Lopez, Gabriela C., Louis D. Van Camp, Ryan F. Kovaleski, et al. “Region-Specific Nucleus Accumbens Dopamine Signals Encode Distinct Aspects of Avoidance Learning.” `Cell Biology`, Volume 35, Issue 10p2433-2443.e5May 19, 2025. DOI: 10.1016/j.cub.2025.04.006
    """

    def __init__(self,
                 data_dir: Union[str, os.PathLike],
                 config_path: Union[str, os.PathLike],
                 nose_name: Optional[str] = 'nose',
                 left_ear_name: Optional[str] = 'Left_ear',
                 right_ear_name: Optional[str] = 'right_ear',
                 tail_base_name: Optional[str] = 'tail_base',
                 time_window: Optional[int] = 3,
                 movement_threshold: Optional[int] = 5,
                 shortest_bout: Optional[int] = 100,
                 save_dir: Optional[Union[str, os.PathLike]] = None):

        check_if_dir_exists(in_dir=data_dir)
        for bp_name in [nose_name, left_ear_name, right_ear_name, tail_base_name]: check_str(name='body part name', value=bp_name, allow_blank=False)
        self.data_paths = find_files_of_filetypes_in_directory(directory=data_dir, extensions=['.csv'])
        ConfigReader.__init__(self, config_path=config_path, read_video_info=True, create_logger=False)
        self.nose_heads = [f'{nose_name}_x'.lower(), f'{nose_name}_y'.lower()]
        self.left_ear_heads = [f'{left_ear_name}_x'.lower(), f'{left_ear_name}_y'.lower()]
        self.right_ear_heads = [f'{right_ear_name}_x'.lower(), f'{right_ear_name}_y'.lower()]
        self.tail_base_heads = [f'{tail_base_name}_x'.lower(), f'{tail_base_name}_y'.lower()]
        self.required_field = self.nose_heads + self.left_ear_heads + self.right_ear_heads + self.tail_base_heads
        check_int(name='time_window', value=time_window, min_value=1)
        check_int(name='movement_threshold', value=movement_threshold, min_value=1)
        self.save_dir = save_dir
        if self.save_dir is None:
            self.save_dir = os.path.join(self.logs_path, f'freezing_data_time_{time_window}s_{self.datetime}')
            os.makedirs(self.save_dir)
        else:
            check_if_dir_exists(in_dir=self.save_dir)
        self.time_window, self.movement_threshold = time_window, movement_threshold
        self.movement_threshold, self.shortest_bout = movement_threshold, shortest_bout
        self.run()

    def run(self):
        agg_results = pd.DataFrame(columns=['VIDEO', 'FREEZING FRAMES', 'FREEZING TIME (S)', 'FREEZING BOUT COUNTS', 'FREEZING PCT OF SESSION', 'VIDEO TOTAL FRAMES', 'VIDEO TOTAL TIME (S)'])
        agg_results_path = os.path.join(self.save_dir, 'aggregate_freezing_results.csv')
        check_all_file_names_are_represented_in_video_log(video_info_df=self.video_info_df, data_paths=self.data_paths)
        for file_cnt, file_path in enumerate(self.data_paths):
            video_name = get_fn_ext(filepath=file_path)[1]
            print(f'Analyzing {video_name}...({file_cnt+1}/{len(self.data_paths)})')
            save_file_path = os.path.join(self.save_dir, f'{video_name}.csv')
            df = read_df(file_path=file_path, file_type='csv').reset_index(drop=True)
            _, px_per_mm, fps = read_video_info(vid_info_df=self.video_info_df, video_name=video_name)
            df.columns = [str(x).lower() for x in df.columns]
            check_valid_dataframe(df=df, valid_dtypes=Formats.NUMERIC_DTYPES.value, required_fields=self.required_field)
            nose_shifted = FeatureExtractionMixin.create_shifted_df(df[self.nose_heads])
            nose_1, nose_2 = nose_shifted.iloc[:, 0:2].values, nose_shifted.iloc[:, 2:4].values
            nose_movement = FeatureExtractionMixin.euclidean_distance(bp_1_x=nose_1[:, 0].flatten(), bp_2_x=nose_2[:, 0].flatten(), bp_1_y=nose_1[:, 1].flatten(), bp_2_y=nose_2[:, 1].flatten(), px_per_mm=px_per_mm)
            tail_base_shifted = FeatureExtractionMixin.create_shifted_df(df[self.tail_base_heads])
            tail_base_shifted_1, tail_base_shifted_2 = tail_base_shifted.iloc[:, 0:2].values, tail_base_shifted.iloc[:, 2:4].values
            tail_base_movement = FeatureExtractionMixin.euclidean_distance(bp_1_x=tail_base_shifted_1[:, 0].flatten(), bp_2_x=tail_base_shifted_2[:, 0].flatten(), bp_1_y=tail_base_shifted_1[:, 1].flatten(), bp_2_y=tail_base_shifted_2[:, 1].flatten(), px_per_mm=px_per_mm)
            left_ear_arr = df[self.left_ear_heads].values.astype(np.int64)
            right_ear_arr = df[self.right_ear_heads].values.astype(np.int64)
            nape_arr = pd.DataFrame(FeatureExtractionMixin.find_midpoints(bp_1=left_ear_arr, bp_2=right_ear_arr, percentile=np.float64(0.5)), columns=[NAPE_X, NAPE_Y])
            nape_shifted = FeatureExtractionMixin.create_shifted_df(nape_arr[[NAPE_X, NAPE_Y]])
            nape_shifted_1, nape_shifted_2 = nape_shifted.iloc[:, 0:2].values, nape_shifted.iloc[:, 2:4].values
            nape_movement = FeatureExtractionMixin.euclidean_distance(bp_1_x=nape_shifted_1[:, 0].flatten(), bp_2_x=nape_shifted_2[:, 0].flatten(), bp_1_y=nape_shifted_1[:, 1].flatten(),  bp_2_y=nape_shifted_2[:, 1].flatten(), px_per_mm=px_per_mm)
            movement = np.hstack([nose_movement.reshape(-1, 1), nape_movement.reshape(-1, 1), tail_base_movement.reshape(-1, 1)])
            mean_movement = np.mean(movement, axis=1)
            mm_s = TimeseriesFeatureMixin.sliding_descriptive_statistics(data=mean_movement.astype(np.float32), window_sizes=np.array([1], dtype=np.float64), sample_rate=int(fps), statistics=typed.List(["sum"]))[0].flatten()
            freezing_idx = np.argwhere(mm_s <= self.movement_threshold).astype(np.int32).flatten()
            df[FREEZING] = 0
            df.loc[freezing_idx, FREEZING] = 1
            df = plug_holes_shortest_bout(data_df=df, clf_name=FREEZING, fps=fps, shortest_bout=self.shortest_bout)
            bouts = detect_bouts(data_df=df, target_lst=[FREEZING], fps=fps)
            bouts = bouts[bouts['Bout_time'] >= self.time_window]
            if len(bouts) > 0:
                freezing_idx = list(bouts.apply(lambda x: list(range(int(x["Start_frame"]), int(x["End_frame"]) + 1)), 1))
                freezing_idx = [x for xs in freezing_idx for x in xs]
                df.loc[freezing_idx, FREEZING] = 1
            else:
                freezing_idx = []
            df.to_csv(save_file_path)
            agg_results.loc[len(agg_results)] = [video_name, len(freezing_idx), round(len(freezing_idx) / fps, 4), len(bouts), round((len(freezing_idx) / len(df)) * 100, 4), len(df), round(len(df)/fps, 2) ]

        agg_results.to_csv(agg_results_path)
        stdout_success(msg=f'Results saved in {self.save_dir} directory.')

#
# FreezingDetector(data_dir=r'C:\troubleshooting\mitra\project_folder\csv\outlier_corrected_movement_location',
#                  config_path=r"C:\troubleshooting\mitra\project_folder\project_config.ini")


__author__ = "Simon Nilsson"

from tkinter import *

import numpy as np
import pandas as pd

from simba.mixins.pop_up_mixin import PopUpMixin
from simba.third_party_label_appenders.transform.dlc_ma_h5_to_yolo import \
    MADLCH52Yolo
from simba.third_party_label_appenders.transform.utils import \
    get_yolo_keypoint_flip_idx
from simba.ui.tkinter_functions import (CreateLabelFrameWithIcon, FolderSelect,
                                        SimbaCheckbox, SimBADropDown)
from simba.utils.checks import check_if_dir_exists, check_valid_dataframe
from simba.utils.enums import Formats, Options
from simba.utils.read_write import (find_files_of_filetypes_in_directory,
                                    str_2_bool)

TRAIN_SIZE_OPTIONS = list(np.arange(10, 110, 10))
SAMPLE_SIZE_OPTIONS = list(np.arange(50, 650, 50))

PADDING_OPTIONS = list(np.round(np.arange(0.01, 10.05, 0.05),2).astype(str))
PADDING_OPTIONS = np.insert(PADDING_OPTIONS, 0, 'None')

THRESHOLD_OPTION = list(range(10, 110, 10))

class DLCH5Inference2YoloPopUp(PopUpMixin):

    """
    GUI for converting DeepLabCut (DLC) annotations to YOLO keypoint format.

    :example:
    >>> DLCH5Inference2YoloPopUp()
    """

    def __init__(self):
        PopUpMixin.__init__(self, title="DLC MULTI-ANIMAL H5 PREDICTIONS TO YOLO", icon='dlc_1')
        settings_frm = CreateLabelFrameWithIcon(parent=self.main_frm, header="SETTINGS", icon_name='settings')
        self.dlc_dir = FolderSelect(settings_frm, folderDescription="DLC DATA DIRECTORY (H5):", lblwidth=35, entry_width=40, initialdir=r'D:\troubleshooting\dlc_h5_multianimal_to_yolo\data')
        self.video_dir = FolderSelect(settings_frm, folderDescription="VIDEO DIRECTORY:", lblwidth=35, entry_width=40, initialdir=r'D:\troubleshooting\dlc_h5_multianimal_to_yolo\videos')
        self.save_dir = FolderSelect(settings_frm, folderDescription="SAVE DIRECTORY:", lblwidth=35, entry_width=40, initialdir='D:\troubleshooting\dlc_h5_multianimal_to_yolo\yolo')

        self.frm_cnt_dropdown = SimBADropDown(parent=settings_frm, dropdown_options=SAMPLE_SIZE_OPTIONS, label="FRAMES (PER VIDEO): ", label_width=35, dropdown_width=40, value=100)
        self.verbose_dropdown = SimBADropDown(parent=settings_frm, dropdown_options=['TRUE', 'FALSE'], label="VERBOSE: ", label_width=35, dropdown_width=40, value='TRUE')
        self.threshold_dropdown = SimBADropDown(parent=settings_frm, dropdown_options=THRESHOLD_OPTION, label="THRESHOLD (%): ", label_width=35, dropdown_width=40, value=90)
        self.train_size_dropdown = SimBADropDown(parent=settings_frm, dropdown_options=TRAIN_SIZE_OPTIONS, label="TRAIN SIZE (%): ", label_width=35, dropdown_width=40, value=70)
        self.grey_dropdown = SimBADropDown(parent=settings_frm, dropdown_options=['TRUE', 'FALSE'], label="GREYSCALE: ", label_width=35, dropdown_width=40, value='FALSE')
        self.clahe_dropdown = SimBADropDown(parent=settings_frm, dropdown_options=['TRUE', 'FALSE'], label="CLAHE: ", label_width=35, dropdown_width=40, value='FALSE')
        self.padding_dropdown = SimBADropDown(parent=settings_frm, dropdown_options=PADDING_OPTIONS, label="PADDING: ", label_width=35, dropdown_width=40, value='None')
        single_id_cb, self.single_id_var = SimbaCheckbox(parent=settings_frm, txt="REMOVE ANIMAL ID'S", val=False)

        settings_frm.grid(row=0, column=0, sticky=NW)
        self.dlc_dir .grid(row=0, column=0, sticky=NW)
        self.video_dir.grid(row=1, column=0, sticky=NW)
        self.save_dir.grid(row=2, column=0, sticky=NW)

        self.frm_cnt_dropdown.grid(row=3, column=0, sticky=NW)
        self.verbose_dropdown.grid(row=4, column=0, sticky=NW)
        self.threshold_dropdown.grid(row=5, column=0, sticky=NW)
        self.train_size_dropdown.grid(row=6, column=0, sticky=NW)
        self.grey_dropdown.grid(row=7, column=0, sticky=NW)
        self.clahe_dropdown.grid(row=8, column=0, sticky=NW)
        self.padding_dropdown.grid(row=9, column=0, sticky=NW)
        single_id_cb.grid(row=10, column=0, sticky=NW)

        self.create_run_frm(run_function=self.run)
        self.main_frm.mainloop()

    def run(self):
        dlc_dir = self.dlc_dir.folder_path
        video_dir = self.video_dir.folder_path
        save_dir = self.save_dir.folder_path

        check_if_dir_exists(in_dir=dlc_dir, source=f'{self.__class__.__name__} SLEAP DATA DIRECTORY', raise_error=True)
        check_if_dir_exists(in_dir=video_dir, source=f'{self.__class__.__name__} VIDEO DIRECTORY', raise_error=True)
        check_if_dir_exists(in_dir=save_dir, source=f'{self.__class__.__name__} SAVE DIRECTORY', raise_error=True)
        h5_files = find_files_of_filetypes_in_directory(directory=dlc_dir, extensions=['.h5', '.H5'], raise_error=True)
        video_files = find_files_of_filetypes_in_directory(directory=video_dir, extensions=Options.ALL_VIDEO_FORMAT_OPTIONS.value, raise_error=True)

        grey = str_2_bool(self.grey_dropdown.get_value())
        clahe = str_2_bool(self.clahe_dropdown.get_value())
        train_size = int(self.train_size_dropdown.get_value()) / 100
        verbose = str_2_bool(self.verbose_dropdown.get_value())
        frm_cnt = int(self.frm_cnt_dropdown.get_value())
        padding = float(self.padding_dropdown.get_value()) if self.padding_dropdown.get_value() != 'None' else 0.0
        threshold = float(self.threshold_dropdown.get_value()) / 100
        single_id = 'animal_1' if self.single_id_var.get() else None

        data_df = pd.DataFrame(pd.read_hdf(h5_files[0], header=[0, 1, 2, 3]))
        check_valid_dataframe(df=data_df, source=h5_files[0], valid_dtypes=Formats.NUMERIC_DTYPES.value)
        animal_names = tuple(list(set([x[1] for x in list(data_df.columns)])))
        bp_names = [x[2] for x in data_df.columns if x[1] == animal_names[0]]
        bp_names = list(dict.fromkeys([x for x in bp_names]))
        flip_idx = get_yolo_keypoint_flip_idx(x=bp_names)

        runner = MADLCH52Yolo(data_dir=dlc_dir,
                              video_dir=video_dir,
                              save_dir=save_dir,
                              frms_cnt=frm_cnt,
                              verbose=verbose,
                              threshold=threshold,
                              train_size=train_size,
                              flip_idx=flip_idx,
                              greyscale=grey,
                              clahe=clahe,
                              padding=padding,
                              single_id=single_id)
        runner.run()

#DLCH5Inference2YoloPopUp()


"""
SimBA (Simple Behavioral Analysis)
https://github.com/sgoldenlab/simba
Contributors.
https://github.com/sgoldenlab/simba#contributors-
Licensed under GNU Lesser General Public License v3.0
"""

import setuptools
import platform
from setuptools import setup, find_namespace_packages

REQUIREMENTS_PATH = 'requirements.txt'
ARM_REQUIREMENTS_PATH = 'requirements_arm.txt'
GPU_REQUIREMENTS_PATH = 'requirements_gpu.txt'

with open("docs/project_description.md", "r", encoding="utf-8") as fh:
    long_description = fh.read()

with open(REQUIREMENTS_PATH, "r") as f:
    requirements = f.read().splitlines()

with open(ARM_REQUIREMENTS_PATH, "r") as f:
    arm_requirements = f.read().splitlines()

with open(GPU_REQUIREMENTS_PATH, "r") as f:
    gpu_requirements = f.read().splitlines()

# Setup configuration
setuptools.setup(
    name="simba_uw_tf_dev",
    version="3.7.1",
    author="Simon Nilsson, Jia Jie Choong, Sophia Hwang",
    author_email="sronilsson@gmail.com",
    description="Toolkit for computer classification and analysis of behaviors in experimental animals",
    long_description=long_description,
    long_description_content_type="text/markdown",
    url="https://github.com/sgoldenlab/simba",
    install_requires=requirements,
    extras_require={'arm': [arm_requirements], "gpu": [gpu_requirements]},
    license='Modified BSD 3-Clause License',
    license_files=('LICENSE',),
    packages=setuptools.find_packages(exclude=["*.tests",
                                               "*.tests.*",
                                               "tests.*",
                                               "tests",
                                               "__pycache__",
                                               "pose_configurations_archive",
                                               "sandbox",
                                               "sandbox.*",
                                               ]),
    # packages=find_namespace_packages(include=["simba*"],
    #                                  exclude=["*.tests", "*.tests.*", "tests.*", "tests", "__pycache__", "__pycache__.*", "pose_configurations_archive", "pose_configurations_archive.*", "sandbox", "sandbox.*"]),
    include_package_data=True,
    classifiers=[
        "Programming Language :: Python :: 3",
        "Operating System :: OS Independent",
    ],
    entry_points={'console_scripts':['simba=simba.SimBA:main'],}
)






__author__ = "Simon Nilsson"

import threading
from tkinter import *

from simba.mixins.config_reader import ConfigReader
from simba.mixins.pop_up_mixin import PopUpMixin
from simba.ui.tkinter_functions import DropDownMenu, FolderSelect
from simba.unsupervised.data_extractor import DataExtractor
from simba.unsupervised.enums import UMLOptions
from simba.utils.checks import check_if_dir_exists
from simba.utils.enums import Formats, Options
from simba.utils.errors import CountError
from simba.utils.read_write import find_files_of_filetypes_in_directory


class DataExtractorPopUp(PopUpMixin, ConfigReader):
    def __init__(self, config_path: str):
        PopUpMixin.__init__(self, title="DATA EXTRACTOR")
        ConfigReader.__init__(self, config_path=config_path)
        data_frm = LabelFrame(
            self.main_frm,
            text="DATA",
            font=Formats.LABELFRAME_HEADER_FORMAT.value,
            fg="black",
        )
        self.data_dir_select = FolderSelect(
            data_frm,
            "DATA DIRECTORY (PICKLES):",
            lblwidth=25,
            initialdir=self.project_path,
        )
        data_frm.grid(row=0, column=0, sticky=NW)
        self.data_dir_select.grid(row=0, column=0, sticky=NW)

        self.cb_settings = self.create_cb_frame(
            main_frm=self.main_frm,
            cb_titles=UMLOptions.DATA_TYPES.value,
            frm_title="DATA TYPES",
        )

        self.create_run_frm(run_function=self.run, title="RUN")
        self.main_frm.mainloop()

    def run(self):
        selections = []
        for k, v in self.cb_settings.items():
            if v.get():
                selections.append(k)
        check_if_dir_exists(self.data_dir_select.folder_path)
        if len(selections) == 0:
            raise CountError(
                msg="Select at least 1 data-type checkbox",
                source=self.__class__.__name__,
            )
        check_if_dir_exists(self.data_dir_select.folder_path)
        _ = find_files_of_filetypes_in_directory(
            directory=self.data_dir_select.folder_path,
            extensions=[f".{Formats.PICKLE.value}"],
            raise_error=True,
        )
        data_extractor = DataExtractor(
            config_path=self.config_path,
            data_path=self.data_dir_select.folder_path,
            data_types=selections,
        )
        threading.Thread(target=data_extractor.run()).start()


# _ = DataExtractorPopUp(config_path='/Users/simon/Desktop/envs/NG_Unsupervised/project_folder/project_config.ini')



import os
import platform
import subprocess
import sys
from datetime import datetime
from tkinter import *
from typing import Union

import numpy as np
from shapely.geometry import MultiPolygon, Point, Polygon

import simba
from simba.mixins.pop_up_mixin import PopUpMixin
from simba.roi_tools.roi_ui import ROI_ui
from simba.roi_tools.roi_utils import get_roi_data, multiply_ROIs
from simba.ui.blob_quick_check_interface import BlobQuickChecker
from simba.ui.tkinter_functions import (CreateLabelFrameWithIcon, FileSelect,
                                        FolderSelect, SimbaButton,
                                        SimBADropDown, SimBALabel)
from simba.utils.checks import (check_if_dir_exists, check_int,
                                check_nvidea_gpu_available, check_str)
from simba.utils.enums import Formats, Paths
from simba.utils.errors import (InvalidInputError, InvalidVideoFileError,
                                NoDataError)
from simba.utils.read_write import (find_all_videos_in_directory,
                                    find_core_cnt,
                                    find_files_of_filetypes_in_directory,
                                    get_video_meta_data, remove_files,
                                    save_json, str_2_bool, write_pickle)
from simba.video_processors.blob_tracking_executor import BlobTrackingExecutor

ABSOLUTE = 'absolute'
DEFAULT_THRESHOLD = 30

WINDOW_SIZES = ['None', 1.0, 1.5, 2.0, 2.5, 3.0, 3.5, 4.0, 4.5, 5.0]
SMOOTHING_TIMES = list(np.round(np.arange(0.01, 2.1, 0.01), 2))
SMOOTHING_TIMES.insert(0, 'None')
BUFFER_SIZES = list(range(1, 101))
BUFFER_SIZES.insert(0, 'None')
KERNEL_SIZES = list(np.round(np.arange(0.1, 100.1, 0.1), 2))
KERNEL_SIZES.insert(0, 'None')

SIMBA_DIR = os.path.dirname(simba.__file__)
BLOB_EXECUTOR_PATH = os.path.join(SIMBA_DIR, Paths.BLOB_EXECUTOR_PATH.value)

class BlobTrackingUI(PopUpMixin):
    """
    :example:
    >>> _ = BlobTrackingUI(input_dir=r'D:\open_field_3\sample', output_dir=r"D:\open_field_3\sample\blob_data")
    """

    def __init__(self,
                 input_dir: Union[str, os.PathLike],
                 output_dir: Union[str, os.PathLike]):

        check_if_dir_exists(in_dir=input_dir, source=self.__class__.__name__)
        check_if_dir_exists(in_dir=output_dir, create_if_not_exist=True, source=self.__class__.__name__)
        if os.path.isdir(output_dir):
            existing_files = find_files_of_filetypes_in_directory(directory=output_dir, extensions=['.json', '.csv', '.mp4', '.pickle'], raise_error=False, raise_warning=False)
            if len(existing_files) > 0: raise InvalidInputError(msg=f'The selected output directory {output_dir} is not empty. Select an empty output directory where to save your tracking data', source=self.__class__.__name__)
        self.gpu_available = check_nvidea_gpu_available()
        self.in_videos = find_all_videos_in_directory(directory=input_dir, as_dict=True, raise_error=True)
        self.input_dir, self.output_dir = input_dir, output_dir
        PopUpMixin.__init__(self, title="BLOB TRACKING", size=(2400, 600), icon='bubble_green')
        self.len_max_char = max(len(max(list(self.in_videos.keys()), key=len)), 25)
        self.core_cnt = find_core_cnt()[0]
        self.get_quick_settings()
        self.get_main_table()
        self.get_main_table_entries()
        self.get_execute_btns()
        self.get_status_bar()
        self.roi_inclusion_store = os.path.join(self.output_dir, 'inclusion_definitions.h5')
        self.out_path = os.path.join(self.output_dir, 'blob_definitions.pickle')
        #self.main_frm.mainloop()

    def get_quick_settings(self):
        self.settings_frm = CreateLabelFrameWithIcon(parent=self.main_frm, header="SETTINGS", icon_name='settings', padx=15, pady=15)
        self.quick_settings_frm = CreateLabelFrameWithIcon(parent=self.settings_frm, header="QUICK SETTINGS", icon_name='clock', padx=15, pady=15, relief='solid')
        self.quick_setting_threshold_dropdown = SimBADropDown(parent=self.quick_settings_frm, dropdown_options=list(range(1, 101)), label="THRESHOLD:", label_width=30, dropdown_width=10, value=DEFAULT_THRESHOLD)
        self.quick_setting_threshold_btn = SimbaButton(parent=self.quick_settings_frm, txt='APPLY', img='tick',cmd=self._set_threshold, cmd_kwargs={'threshold': lambda: self.quick_setting_threshold_dropdown.getChoices()})
        self.set_smoothing_time_dropdown = SimBADropDown(parent=self.quick_settings_frm, dropdown_options=SMOOTHING_TIMES, label="SMOOTHING TIME (S):", label_width=30, dropdown_width=10, value=SMOOTHING_TIMES[0])
        self.set_smoothing_time_btn = SimbaButton(parent=self.quick_settings_frm, txt='APPLY', img='tick', cmd=self._set_smoothing_time, cmd_kwargs={'time': lambda: self.set_smoothing_time_dropdown.getChoices()})
        self.set_buffer_dropdown = SimBADropDown(parent=self.quick_settings_frm, dropdown_options=BUFFER_SIZES, label="BUFFER (PIXELS):", label_width=30, dropdown_width=10, value=BUFFER_SIZES[0])
        self.set_buffer_btn = SimbaButton(parent=self.quick_settings_frm, txt='APPLY', img='tick', cmd=self._set_buffer_size, cmd_kwargs={'size': lambda: self.set_buffer_dropdown.getChoices()})
        self.closing_kernal_size_dropdown = SimBADropDown(parent=self.quick_settings_frm, dropdown_options=KERNEL_SIZES, label="GAP FILL FILTER SIZE (%):", label_width=30, dropdown_width=10, value=BUFFER_SIZES[0])
        self.set_closing_kernel_btn = SimbaButton(parent=self.quick_settings_frm, txt='APPLY', img='tick', cmd=self._set_close_kernel_dropdown, cmd_kwargs={'size': lambda: self.closing_kernal_size_dropdown.getChoices()})
        self.opening_kernal_size_dropdown = SimBADropDown(parent=self.quick_settings_frm, dropdown_options=KERNEL_SIZES, label="NOISE REMOVAL FILTER SIZE (%):", label_width=30, dropdown_width=10, value=BUFFER_SIZES[0])
        self.opening_kernel_btn = SimbaButton(parent=self.quick_settings_frm, txt='APPLY', img='tick', cmd=self._set_open_kernel_dropdown, cmd_kwargs={'size': lambda: self.opening_kernal_size_dropdown.getChoices()})

        self.settings_frm.grid(row=0, column=0, sticky=NW)
        self.quick_settings_frm.grid(row=0, column=0, sticky=NW)
        self.quick_setting_threshold_dropdown.grid(row=0, column=0, sticky=NW)
        self.quick_setting_threshold_btn.grid(row=0, column=1, sticky=NW)
        self.set_smoothing_time_dropdown.grid(row=1, column=0, sticky=NW)
        self.set_smoothing_time_btn.grid(row=1, column=1, sticky=NW)
        self.set_buffer_dropdown.grid(row=4, column=0, sticky=NW)
        self.set_buffer_btn.grid(row=4, column=1, sticky=NW)
        self.closing_kernal_size_dropdown.grid(row=5, column=0, sticky=NW)
        self.set_closing_kernel_btn.grid(row=5, column=1, sticky=NW)
        self.opening_kernal_size_dropdown.grid(row=6, column=0, sticky=NW)
        self.opening_kernel_btn.grid(row=6, column=1, sticky=NW)

        self.run_time_settings_frm = CreateLabelFrameWithIcon(parent=self.settings_frm, header="RUN-TIME SETTINGS", icon_name='run', padx=15, pady=15, relief='solid')
        self.bg_dir = FolderSelect(parent=self.run_time_settings_frm, folderDescription='BACKGROUND DIRECTORY:', lblwidth=30, initialdir=self.input_dir)
        self.bg_dir_apply = SimbaButton(parent=self.run_time_settings_frm, txt='APPLY', img='tick',  cmd=self._apply_bg_dir, cmd_kwargs={'bg_dir': lambda: self.bg_dir.folder_path})
        self.use_gpu_dropdown = SimBADropDown(parent=self.run_time_settings_frm, dropdown_options=['TRUE', 'FALSE'], label="USE GPU:", label_width=30, dropdown_width=10, value='FALSE')
        #if not self.gpu_available: self.use_gpu_dropdown.disable()
        #else: self.use_gpu_dropdown.set_value(value='TRUE')
        self.use_gpu_dropdown.disable()
        self.core_cnt_dropdown = SimBADropDown(parent=self.run_time_settings_frm, dropdown_options=list(range(1, self.core_cnt+1)), label="CPU CORE COUNT:", label_width=30, dropdown_width=10, value=int(self.core_cnt / 2))

        self.vertice_cnt_dropdown = SimBADropDown(parent=self.run_time_settings_frm, dropdown_options=list(range(10, 501)), label="VERTICE COUNT:", label_width=30, dropdown_width=10, value=30)
        self.save_videos_dropdown = SimBADropDown(parent=self.run_time_settings_frm, dropdown_options=['TRUE', 'FALSE'], label="SAVE BACKGROUND VIDEOS:", label_width=30, dropdown_width=10, value='TRUE')
        self.close_iterations_dropdown_dropdown = SimBADropDown(parent=self.run_time_settings_frm, dropdown_options=list(range(1, 20)), label="GAP FILLING ITERATIONS:", label_width=30, dropdown_width=10, value=3)
        self.open_iterations_dropdown = SimBADropDown(parent=self.run_time_settings_frm, dropdown_options=list(range(1, 20)), label="NOISE REMOVAL ITERATIONS:", label_width=30, dropdown_width=10, value=3)
        self.duplicate_inclusion_zones_dropdown = SimBADropDown(parent=self.run_time_settings_frm, dropdown_options=list(self.in_videos.keys()), label="DUPLICATE INCLUSION ZONES:", label_width=30, dropdown_width=self.len_max_char, value=list(self.in_videos.keys())[0])
        self.duplicate_inclusion_zones_btn = SimbaButton(parent=self.run_time_settings_frm, txt='APPLY', img='tick', cmd=self._duplicate_inclusion_zones, cmd_kwargs={'video_name': lambda: self.duplicate_inclusion_zones_dropdown.getChoices()})
        self.run_time_settings_frm.grid(row=0, column=1, sticky=NW)

        self.bg_dir.grid(row=0, column=0, sticky=NW)
        self.bg_dir_apply.grid(row=0, column=1, sticky=NW)
        self.use_gpu_dropdown.grid(row=1, column=0, sticky=NW)
        self.core_cnt_dropdown.grid(row=2, column=0, sticky=NW)
        self.vertice_cnt_dropdown.grid(row=3, column=0, sticky=NW)
        self.save_videos_dropdown.grid(row=4, column=0, sticky=NW)
        self.close_iterations_dropdown_dropdown.grid(row=5, column=0, sticky=NW)
        self.open_iterations_dropdown.grid(row=6, column=0, sticky=NW)
        self.duplicate_inclusion_zones_dropdown.grid(row=7, column=0, sticky=NW)
        self.duplicate_inclusion_zones_btn.grid(row=7, column=1, sticky=NW)
        self.execute_frm = CreateLabelFrameWithIcon(parent=self.settings_frm, header="EXECUTE", icon_name='rocket', padx=15, pady=15, relief='solid')
        self.run_btn = SimbaButton(parent=self.execute_frm, txt='RUN', img='rocket', cmd=self._initialize_run)
        self.remove_inclusion_zones_btn = SimbaButton(parent=self.execute_frm, txt='REMOVE INCLUSION ZONES', img='trash', cmd=self._delete_inclusion_zone_data, cmd_kwargs=None, txt_clr='darkblue')
        self.execute_frm.grid(row=0, column=2, sticky=NW)
        self.run_btn.grid(row=0, column=0, sticky=NW)
        self.remove_inclusion_zones_btn.grid(row=1, column=0, sticky=NW)

    def get_main_table(self):
        self.headings = {}
        self.videos_frm = CreateLabelFrameWithIcon(parent=self.main_frm, header="VIDEOS", icon_name='video', pady=5, padx=15, relief='solid')
        self.headings['video_name'] = SimBALabel(parent=self.videos_frm, txt='VIDEO NAME', width=self.len_max_char, font=Formats.FONT_HEADER.value)
        self.headings['bg_path'] = SimBALabel(parent=self.videos_frm, txt='BACKGROUND REFERENCE', width=25, font=Formats.FONT_HEADER.value)
        self.headings['threshold'] = SimBALabel(parent=self.videos_frm, txt='THRESHOLD', width=15, font=Formats.FONT_HEADER.value)
        self.headings['inclusion_zones'] = SimBALabel(parent=self.videos_frm, txt='INCLUSION ZONES', width=25, font=Formats.FONT_HEADER.value)
        self.headings['smoothing_time'] = SimBALabel(parent=self.videos_frm, txt='SMOOTHING TIME (S)', width=25, font=Formats.FONT_HEADER.value)
        self.headings['buffer_size'] = SimBALabel(parent=self.videos_frm, txt='BUFFER SIZE (PX)', width=25, font=Formats.FONT_HEADER.value)
        self.headings['closing_kernel_size'] = SimBALabel(parent=self.videos_frm, txt='GAP FILL SIZE (%)', width=25, font=Formats.FONT_HEADER.value)
        self.headings['opening_kernel_size'] = SimBALabel(parent=self.videos_frm, txt='NOISE FILL SIZE (%)', width=25, font=Formats.FONT_HEADER.value)
        self.headings['quick_check'] = SimBALabel(parent=self.videos_frm, txt='QUICK CHECK', width=20, font=Formats.FONT_HEADER.value)
        for cnt, (k, v) in enumerate(self.headings.items()):
            v.grid(row=0, column=cnt, sticky=NW)
        self.videos_frm.grid(row=1, column=0, sticky=NW)

    def get_main_table_entries(self):
        self.videos = {}
        for video_cnt, (video_name, video_path) in enumerate(self.in_videos.items()):
            self.videos[video_name] = {}
            self.videos[video_name]['name_lbl'] = SimBALabel(parent=self.videos_frm, txt=video_name, width=self.len_max_char, font=Formats.FONT_HEADER.value)
            self.videos[video_name]["threshold_dropdown"] = SimBADropDown(parent=self.videos_frm, dropdown_options=list(range(1, 100)), label="", label_width=0, dropdown_width=15, value=DEFAULT_THRESHOLD)
            self.videos[video_name]["inclusion_btn"] = Button(self.videos_frm, text="SET INCLUSION ZONES", fg="black", command=lambda k=self.videos[video_name]['name_lbl']["text"]: self._launch_set_inclusion_zones(k), width=25)
            self.videos[video_name]["bg_file"] = FileSelect(parent=self.videos_frm, width=25)
            self.videos[video_name]["smoothing_time_dropdown"] = SimBADropDown(parent=self.videos_frm, dropdown_options=SMOOTHING_TIMES, label="", label_width=0, dropdown_width=25, value=SMOOTHING_TIMES[0])
            self.videos[video_name]["quick_check_btn"] = Button(self.videos_frm, text="QUICK CHECK", fg="black", command=lambda k=self.videos[video_name]['name_lbl']["text"]: self._quick_check(k), width=20)
            self.videos[video_name]["buffer_dropdown"] = SimBADropDown(parent=self.videos_frm, dropdown_options=BUFFER_SIZES, label="", label_width=0, dropdown_width=25, value=BUFFER_SIZES[0])
            self.videos[video_name]["close_kernel"] = SimBADropDown(parent=self.videos_frm, dropdown_options=KERNEL_SIZES, label="", label_width=0, dropdown_width=25, value=KERNEL_SIZES[0])
            self.videos[video_name]["open_kernel"] = SimBADropDown(parent=self.videos_frm, dropdown_options=KERNEL_SIZES, label="", label_width=0, dropdown_width=25, value=KERNEL_SIZES[0])
            self.videos[video_name]['name_lbl'].grid(row=video_cnt+1, column=0, sticky=NW)
            self.videos[video_name]['bg_file'].grid(row=video_cnt+1, column=1, sticky=NW)
            self.videos[video_name]['threshold_dropdown'].grid(row=video_cnt + 1, column=2, sticky=NW)
            self.videos[video_name]['inclusion_btn'].grid(row=video_cnt+1, column=3, sticky=NW)
            self.videos[video_name]['smoothing_time_dropdown'].grid(row=video_cnt + 1, column=4, sticky=NW)
            self.videos[video_name]['buffer_dropdown'].grid(row=video_cnt + 1, column=5, sticky=NW)
            self.videos[video_name]['close_kernel'].grid(row=video_cnt + 1, column=6, sticky=NW)
            self.videos[video_name]['open_kernel'].grid(row=video_cnt + 1, column=7, sticky=NW)
            self.videos[video_name]['quick_check_btn'].grid(row=video_cnt + 1, column=8, sticky=NW)

    def get_execute_btns(self):
        self.execute_frm = CreateLabelFrameWithIcon(parent=self.main_frm, header="EXECUTE", icon_name='rocket', padx=15, pady=15, relief='solid')
        self.run_btn = SimbaButton(parent=self.execute_frm, txt='RUN', img='rocket', cmd=self._initialize_run)
        self.remove_inclusion_zones_btn = SimbaButton(parent=self.execute_frm, txt='REMOVE INCLUSION ZONES', img='trash', cmd=self._delete_inclusion_zone_data, cmd_kwargs=None, txt_clr='darkblue')
        self.execute_frm.grid(row=2, column=0, sticky=NW)
        self.run_btn.grid(row=0, column=0, sticky=NW)
        self.remove_inclusion_zones_btn.grid(row=0, column=1, sticky=NW)

    def get_status_bar(self):
        self.status_bar = Label(self.main_frm, text="STATUS: READY", bd=1, relief=SUNKEN, anchor=W, fg='blue')
        self.status_bar.grid(row=3, column=0, sticky="ew")

    def set_status_bar_panel(self, text: str, fg: str = 'blue'):
        self.status_bar.configure(text=text, fg=fg)
        self.status_bar.update_idletasks()
        print(text)


    def _delete_inclusion_zone_data(self):
        if os.path.isfile(self.roi_inclusion_store):
            remove_files(file_paths=[self.roi_inclusion_store])
        else:
            msg = 'Cannot delete INCLUSION zones: No INCLUSION zones found.'
            self.set_status_bar_panel(text=msg, fg='red')
            raise NoDataError(msg='Cannot delete INCLUSION zones: No INCLUSION zones exist.', source=self.__class__.__name__)

    def _set_window_size(self, size: str):
        for video_name in self.videos.keys():
            self.videos[video_name]["window_size_dropdown"].setChoices(size)

    def _launch_set_inclusion_zones(self, video_name: str):
        ROI_ui(config_path=None, video_path=self.in_videos[video_name], roi_coordinates_path=self.roi_inclusion_store, video_dir=self.input_dir)


    def _set_smoothing_time(self, time: str):
        for video_name in self.videos.keys():
            self.videos[video_name]["smoothing_time_dropdown"].setChoices(time)
        self.set_status_bar_panel(text=f'Set smoothing time to {time} for all videos...', fg='blue')

    def _set_buffer_size(self, size: str):
        for video_name in self.videos.keys():
            self.videos[video_name]["buffer_dropdown"].setChoices(size)
        self.set_status_bar_panel(text=f'Set buffer size to {size} for all videos...', fg='blue')

    def _set_close_kernel_dropdown(self, size: str):
        for video_name in self.videos.keys():
            self.videos[video_name]["close_kernel"].setChoices(size)
        self.set_status_bar_panel(text=f'Set CLOSE kernel size to {size} for all videos...', fg='blue')


    def _set_open_kernel_dropdown(self, size: str):
        for video_name in self.videos.keys():
            self.videos[video_name]["open_kernel"].setChoices(size)
        self.set_status_bar_panel(text=f'Set OPEN kernel size to {size} for all videos...', fg='blue')


    def _set_threshold(self, threshold: str):
        check_int(name='threshold', value=threshold, min_value=0, max_value=100, raise_error=True)
        for video_name in self.videos.keys():
            self.videos[video_name]["threshold_dropdown"].setChoices(threshold)
        self.set_status_bar_panel(text=f'Set threshold to {threshold} for all videos...', fg='blue')

    def _set_visualize(self, val: str):
        check_str(name='visualize', value=val, options=('TRUE', 'FALSE'), raise_error=True)
        for video_name in self.videos.keys():
            self.videos[video_name]["visualize_dropdown"].setChoices(val)

    def _duplicate_inclusion_zones(self, video_name: str):
        if os.path.isfile(self.roi_inclusion_store):
            _, _, _, roi_dict, _, _, _ = get_roi_data(roi_path=self.roi_inclusion_store, video_name=video_name)
            if len(list(roi_dict.keys())) == 0:
                msg = f'Cannot duplicate the INCLUSION zones in {video_name}: Video {video_name} have no drawn INCLUSION zones.'
                self.set_status_bar_panel(text=msg, fg='red')
                raise NoDataError(msg=msg, source=self.__class__.__name__)
            else:
                multiply_ROIs(filename=self.in_videos[video_name], roi_coordinates_path=self.roi_inclusion_store, videos_dir=self.input_dir)
                self.set_status_bar_panel(text=f'Duplicated inclusion zoned for {video_name} for all videos...', fg='blue')
        else:
            msg = f'Cannot duplicate the INCLUSION zones in {video_name}: No INCLUSION zones have been drawn.'
            self.set_status_bar_panel(text=msg, fg='red')
            raise NoDataError(msg=msg, source=self.__class__.__name__)


    def _check_bg_videos(self, bg_videos: dict, videos: dict):
        missing_bg_videos = list(set(list(videos.keys())) - set(list(bg_videos.keys())))
        if len(missing_bg_videos) > 0:
            msg = f'The chosen BACKGROUND DIRECTORY is missing videos for {len(missing_bg_videos)} video files: {missing_bg_videos}'
            self.set_status_bar_panel(text=msg, fg='red')
            raise InvalidVideoFileError(msg=msg, source=self.__class__.__name__)
        for video_name, video_path in self.in_videos.items():
            video_meta = get_video_meta_data(video_path=video_path)
            bg_meta = get_video_meta_data(video_path=bg_videos[video_name])
            if video_meta['resolution_str'] != bg_meta['resolution_str']:
                msg = f'The video, and background reference video, for {video_name} have different resolutions: {video_meta["resolution_str"]} vs {bg_meta["resolution_str"]}'
                self.set_status_bar_panel(text=msg, fg='red')
                raise InvalidVideoFileError(msg=msg, source=self.__class__.__name__)

    def _apply_bg_dir(self, bg_dir: Union[str, os.PathLike]):
        check_if_dir_exists(in_dir=bg_dir)
        self.bg_videos = find_all_videos_in_directory(directory=bg_dir, as_dict=True, raise_error=False)
        self._check_bg_videos(bg_videos=self.bg_videos, videos=self.videos)
        for video_name, video_path in self.in_videos.items():
            self.videos[video_name]['bg_file'].filePath.set(self.bg_videos[video_name])

    def _get_bg_videos(self) -> dict:
        bg_videos = {}
        for video_name in self.videos.keys():
            if not os.path.isfile(self.videos[video_name]['bg_file'].file_path):
                msg = f'The background reference file for {video_name} is not a valid file: {self.videos[video_name]["bg_file"].file_path}.'
                self.set_status_bar_panel(text=msg, fg='red')
                raise InvalidVideoFileError(msg=msg, source=self.__class__.__name__)
            bg_videos[video_name] = self.videos[video_name]['bg_file'].file_path
        self._check_bg_videos(bg_videos=bg_videos, videos=self.in_videos)
        return bg_videos

    def get_roi_definitions(self, video_name: str, coordinates_path: Union[str, os.PathLike]) -> Union[MultiPolygon, None]:
        """ Convert all ROI definitions for specified video into a single multipolygon """
        if os.path.isfile(coordinates_path):
            _, _, _, roi_dict, _, _, _ = get_roi_data(roi_path=coordinates_path, video_name=video_name)
            polygons = []
            for _, roi_data in roi_dict.items():
                if roi_data['Shape_type'] != 'circle':
                    tags = np.array(list(roi_data['Tags'].values()))
                    polygons.append(Polygon(tags))
                else:
                    circle = Point((roi_data['centerX'], roi_data['centerY'])).buffer(roi_data['radius'])
                    polygons.append(Polygon(circle))
            return MultiPolygon(polygons)
        else:
            return None


    def _quick_check(self, video_name: str):
        video_path = self.in_videos[video_name]
        bg_video_path = self.videos[video_name]["bg_file"].file_path
        threshold = int((int(self.videos[video_name]["threshold_dropdown"].getChoices()) / 100) * 255)
        inclusion_zones = self.get_roi_definitions(video_name=video_name, coordinates_path=self.roi_inclusion_store)
        video_meta_data = get_video_meta_data(video_path=video_path)
        close_kernel_iterations = int(self.close_iterations_dropdown_dropdown.get_value())
        open_kernel_iterations = int(self.open_iterations_dropdown.get_value())
        max_dim = max(video_meta_data['width'], video_meta_data['height'])
        if self.videos[video_name]['close_kernel'].get_value() == 'None':
            close_kernel = None
        else:
            w = ((max_dim * float(self.videos[video_name]['close_kernel'].get_value())) / 100) / 5
            h = ((max_dim * float(self.videos[video_name]['close_kernel'].get_value())) / 100) / 5
            close_kernel = (int(max(h, 1)), int(max(w, 1)))
        if self.videos[video_name]['open_kernel'].get_value() == 'None':
            open_kernel = None
        else:
            w = ((max_dim * float(self.videos[video_name]['open_kernel'].get_value())) / 100) / 5
            h = ((max_dim * float(self.videos[video_name]['open_kernel'].get_value())) / 100) / 5
            open_kernel = (int(max(h, 1)), int(max(w, 1)))

        if not os.path.isfile(bg_video_path):
            msg = f'The selected background video selected for {video_name} does not exist'
            self.set_status_bar_panel(text=msg, fg='red')
            raise InvalidVideoFileError(msg=msg, source=self.__class__.__name__)
        else:
            _ = BlobQuickChecker(video_path=video_path,
                                 bg_video_path=bg_video_path,
                                 threshold=threshold,
                                 method='absolute',
                                 inclusion_zones=inclusion_zones,
                                 status_label=self.status_bar,
                                 close_kernel_size=close_kernel,
                                 close_kernel_iterations=close_kernel_iterations,
                                 open_kernel_size=open_kernel,
                                 open_kernel_iterations=open_kernel_iterations)

    def _initialize_run(self):
        _ = self._get_bg_videos()
        out = {'input_dir':  self.input_dir,
               'output_dir': self.output_dir,
               'gpu': str_2_bool(self.use_gpu_dropdown.getChoices()),
               'core_cnt': int(self.core_cnt_dropdown.getChoices()),
               'vertice_cnt': int(self.vertice_cnt_dropdown.get_value()),
               'close_iterations': int(self.close_iterations_dropdown_dropdown.get_value()),
               'open_iterations': int(self.open_iterations_dropdown.get_value()),
               'save_bg_videos': str_2_bool(self.save_videos_dropdown.getChoices())}

        video_out = {}
        for video_name, video_data in self.videos.items():
            video_meta_data = get_video_meta_data(video_path=self.in_videos[video_name])
            video_out[video_name] = {}
            video_out[video_name]['video_path'] = self.in_videos[video_name]
            video_out[video_name]['threshold'] = int((int(self.videos[video_name]["threshold_dropdown"].getChoices()) / 100) * 255)
            video_out[video_name]['smoothing_time'] = self.videos[video_name]["smoothing_time_dropdown"].getChoices()
            video_out[video_name]['buffer_size'] = self.videos[video_name]["buffer_dropdown"].getChoices()
            video_out[video_name]['reference'] = self.videos[video_name]["bg_file"].file_path
            video_out[video_name]['inclusion_zones'] = self.get_roi_definitions(video_name=video_name, coordinates_path=self.roi_inclusion_store)
            video_out[video_name]['window_size'] = None
            video_out[video_name]['close_kernel'] = self.videos[video_name]["close_kernel"].get_value()
            video_out[video_name]['open_kernel'] = self.videos[video_name]["open_kernel"].get_value()
            max_dim = max(video_meta_data['width'], video_meta_data['height'])
            if video_out[video_name]['smoothing_time'] == 'None':
                video_out[video_name]['smoothing_time'] = None
            else:
                video_out[video_name]['smoothing_time'] = int(float(video_out[video_name]['smoothing_time']) * 1000)
            if video_out[video_name]['buffer_size'] == 'None':
                video_out[video_name]['buffer_size'] = None
            else:
                video_out[video_name]['buffer_size'] = int(video_out[video_name]['buffer_size'])
            if video_out[video_name]['close_kernel'] == 'None':
                video_out[video_name]['close_kernel'] = None
            else:
                w = ((max_dim * float(video_out[video_name]['close_kernel'])) / 100) / 4
                h = ((max_dim * float(video_out[video_name]['close_kernel'])) / 100) / 4
                k = (int(max(h, 1)), int(max(w, 1)))
                video_out[video_name]['close_kernel'] = tuple(k)
            if video_out[video_name]['open_kernel'] == 'None':
                video_out[video_name]['open_kernel'] = None
            else:
                w = ((max_dim * float(video_out[video_name]['open_kernel'])) / 100) / 4
                h = ((max_dim * float(video_out[video_name]['open_kernel'])) / 100) / 4
                k = (int(max(h, 1)), int(max(w, 1)))
                video_out[video_name]['open_kernel'] = tuple(k)

        remove_files(file_paths=[self.roi_inclusion_store], raise_error=False)
        out['video_data'] = video_out
        write_pickle(data=out, save_path=self.out_path)
        self.set_status_bar_panel(text=f'Starting blob detection, follow progress in OS terminal...({datetime.now().strftime("%H:%M:%S")})', fg='blue')

        sp = False if platform.system() != "Darwin" else True
        print(f'SimBA blob tracking subprocess: {sp} (platform: {platform.system()}).')
        if sp:
            cmd = f'python "{BLOB_EXECUTOR_PATH}" --data "{self.out_path}"'
            subprocess.run(cmd, check=True, shell=True)
        else:
            executor = BlobTrackingExecutor(data=self.out_path)
            executor.run()

#_ = BlobTrackingUI(input_dir=r'C:\troubleshooting\blob_track_tester\videos', output_dir=r'C:\troubleshooting\blob_track_tester\results')

#_ = BlobTrackingUI(input_dir=r'D:\water_maze', output_dir=r'D:\water_maze\data')


#_ = BlobTrackingUI(input_dir=r'D:\water_maze', output_dir=r'D:\water_maze\data')

#_ = BlobTrackingUI(input_dir=r'D:\OF_7', output_dir=r'D:\OF_7\data')

#_ = BlobTrackingUI(input_dir=r'/mnt/d/open_field_3/sample', output_dir=r"/mnt/d/open_field_3/sample/blob_data")
#_ = BlobTrackingUI(input_dir=r'D:\EPM\sample_2', output_dir=r"D:\EPM\sample_2\data_2")



#_ = BlobTrackingUI(input_dir=r'D:\open_field', output_dir=r"D:\open_field\data")




#_ = BlobTrackingUI(input_dir=r'D:\EPM\sampled', output_dir=r"D:\EPM\sampled\data")

#_ = BlobTrackingUI(input_dir=r'D:\open_field', output_dir=r"D:\open_field\data")

#_ = BlobTrackingUI(input_dir=r'D:\open_field_3\sample', output_dir=r"D:\open_field_3\sample\blob_data")
# import numpy as np
# from shapely.geometry import MultiPolygon, Polygon, Point
#
# rois = {'inclusion_zones': {'rectangle': {'Video': '501_MA142_Gi_Saline_0515', 'Shape_type': 'rectangle', 'Name': 'rectangle', 'Color name': 'Red', 'Color BGR': (0, 0, 255), 'Thickness': 5, 'Center_X': 358, 'Center_Y': 150, 'topLeftX': 195, 'topLeftY': 108, 'Bottom_right_X': 522, 'Bottom_right_Y': 193, 'width': 327, 'height': 85, 'width_cm': 32.7, 'height_cm': 8.5, 'area_cm': 277.95, 'Tags': {'Center tag': (358, 150), 'Top left tag': (195, 108), 'Bottom right tag': (522, 193), 'Top right tag': (522, 108), 'Bottom left tag': (195, 193), 'Top tag': (358, 108), 'Right tag': (522, 150), 'Left tag': (195, 150), 'Bottom tag': (358, 193)}, 'Ear_tag_size': 15}, 'circle': {'Video': '501_MA142_Gi_Saline_0515', 'Shape_type': 'circle', 'Name': 'circle', 'Color name': 'Red', 'Color BGR': (0, 0, 255), 'Thickness': 5, 'centerX': 354, 'centerY': 349, 'radius': 79, 'radius_cm': 7.9, 'area_cm': 196.07, 'Tags': {'Center tag': (354, 349), 'Border tag': (275, 349)}, 'Ear_tag_size': 15}, 'polygon': {'Video': '501_MA142_Gi_Saline_0515', 'Shape_type': 'polygon', 'Name': 'polygon', 'Color name': 'Red', 'Color BGR': (0, 0, 255), 'Thickness': 5, 'Center_X': 571, 'Center_Y': 379, 'vertices': np.array([[676, 286],
#        [643, 462],
#        [474, 496],
#        [504, 276]]), 'center': (571, 379), 'area': 33381.0, 'max_vertice_distance': 291, 'area_cm': 33381.0, 'Tags': {'Tag_0': (676, 286), 'Tag_1': (643, 462), 'Tag_2': (474, 496), 'Tag_3': (504, 276), 'Center_tag': (571, 379)}, 'Ear_tag_size': 15}}}
#
# inclusion_zones = rois['inclusion_zones']
# all_tags, polygons = [], []
# for zone_name, zone_data in inclusion_zones.items():
#     if zone_data['Shape_type'] != 'circle':
#         tags = zone_data['Tags']
#         tags = np.array(list(tags.values()))
#         polygons.append(Polygon(tags))
#     else:
#         circle = Point((zone_data['centerX'], zone_data['centerY'])).buffer(zone_data['radius'])
#         polygons.append(Polygon(circle))
#     multi_polygon = MultiPolygon(polygons)


#
# multipolygon = MultiPolygon(all_tags)







# tags = {'Tags': {'Center tag': (365, 293), 'Top left tag': (204, 131), 'Bottom right tag': (526, 455), 'Top right tag': (526, 131), 'Bottom left tag': (204, 455), 'Top tag': (365, 131), 'Right tag': (526, 293), 'Left tag': (204, 293), 'Bottom tag': (365, 455)}}
#
# import numpy as np
#


import math
import os
import warnings
from copy import copy, deepcopy
from tkinter import *
from typing import Dict, List, Optional, Tuple, Union

import cv2
import numpy as np
import pandas as pd
from PIL import ImageTk
from scipy.spatial.distance import cdist
from shapely.geometry import Polygon

from simba.utils.checks import (check_file_exist_and_readable, check_int,
                                check_str, check_valid_array,
                                check_valid_dataframe, check_valid_tuple,
                                check_video_and_data_frm_count_align)
from simba.utils.enums import (ROI_SETTINGS, ConfigKey, Formats, Keys, Options,
                               Paths)
from simba.utils.errors import (InvalidInputError, NoROIDataError,
                                NotDirectoryError)
from simba.utils.printing import stdout_success, stdout_trash
from simba.utils.read_write import (find_files_of_filetypes_in_directory,
                                    get_fn_ext, get_video_meta_data,
                                    read_config_file, read_df, read_roi_data)
from simba.utils.warnings import (FrameRangeWarning, NoFileFoundWarning,
                                  VideoFileWarning)
from simba.video_processors.roi_selector import ROISelector
from simba.video_processors.roi_selector_circle import ROISelectorCircle
from simba.video_processors.roi_selector_polygon import ROISelectorPolygon

warnings.simplefilter(action='ignore', category=pd.errors.PerformanceWarning)

DRAW_FRAME_NAME = "DEFINE SHAPE"

def create_rectangle_entry(rectangle_selector: ROISelector, video_name: str, shape_name: str, clr_name: str, clr_bgr: Tuple[int, int, int], thickness: int, ear_tag_size: int, px_conversion_factor: float):
    return {'Video':             video_name,
            'Shape_type':        ROI_SETTINGS.RECTANGLE.value,
            'Name':              shape_name,
            'Color name':        clr_name,
            'Color BGR':         clr_bgr,
            'Thickness':         thickness,
            'Center_X':          rectangle_selector.center[0],
            'Center_Y':          rectangle_selector.center[1],
            'topLeftX':          rectangle_selector.top_left[0],
            'topLeftY':          rectangle_selector.top_left[1],
            'Bottom_right_X':    rectangle_selector.bottom_right[0],
            'Bottom_right_Y':    rectangle_selector.bottom_right[1],
            'width':             rectangle_selector.width,
            'height':            rectangle_selector.height,
            'width_cm':          round((rectangle_selector.width / px_conversion_factor) / 10, 2),
            'height_cm':         round((rectangle_selector.height / px_conversion_factor) / 10, 2),
            'area_cm':           round((round((rectangle_selector.width / px_conversion_factor) / 10, 2) * round((rectangle_selector.height / px_conversion_factor) / 10, 2)), 2),
            "Tags":             {"Center tag": rectangle_selector.center,
                                 "Top left tag": rectangle_selector.top_left,
                                 "Bottom right tag": rectangle_selector.bottom_right,
                                 "Top right tag": rectangle_selector.top_right_tag,
                                 "Bottom left tag": rectangle_selector.bottom_left_tag,
                                 "Top tag": rectangle_selector.top_tag,
                                 "Right tag": rectangle_selector.right_tag,
                                 "Left tag": rectangle_selector.left_tag,
                                 "Bottom tag": rectangle_selector.bottom_tag},
            'Ear_tag_size':      ear_tag_size}

def create_circle_entry(circle_selector: ROISelectorCircle,  video_name: str, shape_name: str, clr_name: str, clr_bgr: Tuple[int, int, int], thickness: int, ear_tag_size: int, px_conversion_factor: float):
    return {'Video':             video_name,
            'Shape_type':        ROI_SETTINGS.CIRCLE.value,
            'Name':              shape_name,
            'Color name':        clr_name,
            'Color BGR':         clr_bgr,
            'Thickness':         thickness,
            'centerX':           circle_selector.circle_center[0],
            'centerY':           circle_selector.circle_center[1],
            'radius':            circle_selector.circle_radius,
            'radius_cm':         round((circle_selector.circle_radius / px_conversion_factor) / 10, 2),
            'area_cm':           round(math.pi * (round((circle_selector.circle_radius / px_conversion_factor) / 10, 2) **2), 2),
            "Tags":             {"Center tag": circle_selector.circle_center,
                                 "Border tag": circle_selector.left_border_tag},
            'Ear_tag_size':      ear_tag_size}


def create_polygon_entry(polygon_selector: ROISelectorPolygon, video_name: str, shape_name: str, clr_name: str, clr_bgr: Tuple[int, int, int], thickness: int, ear_tag_size: int, px_conversion_factor: float) -> dict:
    return {'Video':                video_name,
            'Shape_type':           ROI_SETTINGS.POLYGON.value,
            'Name':                 shape_name,
            'Color name':           clr_name,
            'Color BGR':            clr_bgr,
            'Thickness':            thickness,
            'Center_X':             polygon_selector.polygon_centroid[0],
            'Center_Y':             polygon_selector.polygon_centroid[1],
            'vertices':             polygon_selector.polygon_arr,
            'center':               tuple(polygon_selector.polygon_centroid),
            'area':                 polygon_selector.polygon_area,
            'max_vertice_distance': polygon_selector.max_vertice_distance,
            "area_cm":              round(polygon_selector.polygon_area / px_conversion_factor, 2),
            'Tags':                 polygon_selector.tags,
            'Ear_tag_size':         ear_tag_size}

def create_duplicated_rectangle_entry(shape_entry: dict, jump_size: int) -> dict:
    shape_entry['Name'] = f'{shape_entry["Name"]}_duplicated'
    shape_entry['Center_X'] = shape_entry['Center_X'] + jump_size
    shape_entry['Center_Y'] = shape_entry['Center_Y'] + jump_size
    shape_entry['topLeftX'] = shape_entry['topLeftX'] + jump_size
    shape_entry['topLeftY'] = shape_entry['topLeftY'] + jump_size
    shape_entry['Bottom_right_X'] = shape_entry['Bottom_right_X'] + jump_size
    shape_entry['Bottom_right_Y'] = shape_entry['Bottom_right_Y'] + jump_size
    new_shape_tags = {}
    for tag_name, tag_value in shape_entry['Tags'].items():
        new_shape_tags[tag_name] = (int(tag_value[0] + jump_size), int(tag_value[1] + jump_size))
    shape_entry['Tags'] = new_shape_tags
    return shape_entry


def create_duplicated_circle_entry(shape_entry: dict, jump_size: int) -> dict:
    out = copy(shape_entry)
    out['Name'] = f'{out["Name"]}_duplicated'
    out['centerX'] = out['centerX'] + jump_size
    out['centerY'] = out['centerY'] + jump_size
    out['Tags'] = {'Center tag': ((out['centerX']), int(out['centerY'])),
                   'Border tag': (int(out['Tags']['Border tag'][0] + jump_size), int(out['Tags']['Border tag'][1] + jump_size))}
    return out

def create_duplicated_polygon_entry(shape_entry: dict, jump_size: int) -> dict:
    shape_entry['Name'] = f'{shape_entry["Name"]}_duplicated'
    shape_entry['Center_X'] = shape_entry['Center_X'] + jump_size
    shape_entry['Center_Y'] = shape_entry['Center_Y'] + jump_size
    shape_entry['vertices'] = shape_entry['vertices'] + jump_size
    new_shape_tags = {}
    for tag_name, tag_value in shape_entry['Tags'].items():
        new_shape_tags[tag_name] = (int(tag_value[0] + jump_size), int(tag_value[1] + jump_size))
    shape_entry['Tags'] = new_shape_tags
    return shape_entry

def get_rectangle_df_headers():
    return ['Video', 'Shape_type', 'Name', 'Color name', 'Color BGR', 'Thickness', 'Center_X', 'Center_Y', 'topLeftX', 'topLeftY', 'Bottom_right_X', 'Bottom_right_Y', 'width', 'height', 'width_cm', 'height_cm', 'area_cm', "Tags", 'Ear_tag_size']

def get_circle_df_headers():
    return ['Video', 'Shape_type', 'Name', 'Color name', 'Color BGR', 'Thickness', 'centerX', 'centerY', 'radius', 'radius_cm', 'area_cm', "Tags", 'Ear_tag_size']

def get_polygon_df_headers():
    return ['Video', 'Shape_type', 'Name', 'Color name', 'Color BGR', 'Thickness', 'Center_X', 'Center_Y', 'vertices', 'center', 'area', 'max_vertice_distance', 'area_cm', "Tags", 'Ear_tag_size']


def set_roi_metric_sizes(roi_dict: dict, px_conversion_factor: Union[int, float]) -> dict:
    """
    Helper to update the metric attributes of a shape dictionary.
    """
    out = {}
    for name, data in roi_dict.items():
        if data['Shape_type'].lower() == ROI_SETTINGS.RECTANGLE.value:
            data['width_cm'] = round((data['width'] / px_conversion_factor) / 10, 2)
            data['height_cm'] = round((data['height'] / px_conversion_factor) / 10, 2)
            data['area_cm'] =  round((round((data['width'] / px_conversion_factor) / 10, 2) * round((data['height'] / px_conversion_factor) / 10, 2)), 2)
        elif data['Shape_type'].lower() == ROI_SETTINGS.CIRCLE.value:
            data['radius_cm'] = round((data['radius'] / px_conversion_factor) / 10, 2)
            data['area_cm'] = round(math.pi * (round((data['radius'] / px_conversion_factor) / 10, 2) ** 2), 2)
        elif data['Shape_type'].lower() == ROI_SETTINGS.POLYGON.value:
            data['area'] = Polygon(data['vertices']).simplify(tolerance=20, preserve_topology=True).area
            data['area_cm'] = round(data['area'] / px_conversion_factor, 2)
            data['max_vertice_distance'] =  np.max(cdist(data['vertices'], data['vertices']).astype(np.int32))
        out[name] = data
    return out


def get_roi_df_from_dict(roi_dict: dict, video_name_nesting: Optional[bool] = False) -> Tuple[pd.DataFrame, pd.DataFrame, pd.DataFrame]:
    """
    Helper create DataFrames from a shape dictionary.

    If nesting is True, then the roi_dict has video name, and shape name keys.
    """
    rectangles_df, circles_df, polygon_df = pd.DataFrame(columns=get_rectangle_df_headers()), pd.DataFrame(columns=get_circle_df_headers()), pd.DataFrame(columns=get_polygon_df_headers())
    if not video_name_nesting:
        for shape_name, shape_data in roi_dict.items():
            if shape_data['Shape_type'].lower() == ROI_SETTINGS.RECTANGLE.value:
                rectangles_df = pd.concat([rectangles_df, pd.DataFrame([shape_data])], ignore_index=True)
            elif shape_data['Shape_type'].lower() == ROI_SETTINGS.CIRCLE.value:
                circles_df = pd.concat([circles_df, pd.DataFrame([shape_data])], ignore_index=True)
            elif shape_data['Shape_type'].lower() == ROI_SETTINGS.POLYGON.value:
                polygon_df = pd.concat([polygon_df, pd.DataFrame([shape_data])], ignore_index=True)
    else:
        for video_name, video_data in roi_dict.items():
            for shape_name, shape_data in video_data.items():
                if shape_data['Shape_type'].lower() == ROI_SETTINGS.RECTANGLE.value:
                    rectangles_df = pd.concat([rectangles_df, pd.DataFrame([shape_data])], ignore_index=True)
                elif shape_data['Shape_type'].lower() == ROI_SETTINGS.CIRCLE.value:
                    circles_df = pd.concat([circles_df, pd.DataFrame([shape_data])], ignore_index=True)
                elif shape_data['Shape_type'].lower() == ROI_SETTINGS.POLYGON.value:
                    polygon_df = pd.concat([polygon_df, pd.DataFrame([shape_data])], ignore_index=True)

    return (rectangles_df, circles_df, polygon_df)


def get_roi_dict_from_dfs(rectangle_df: pd.DataFrame,
                          circle_df: pd.DataFrame,
                          polygon_df: pd.DataFrame,
                          video_name_nesting: Optional[bool] = False) -> dict:
    """
    Helper create dict from a shape dataframes.
    """

    out = {}
    for idx, row in rectangle_df.iterrows():
        if not video_name_nesting:
            out[row['Name']] = row.to_dict()
        else:
            if row['Video'] not in list(out.keys()):
                out[row['Video']] = {}
            out[row['Video']][row['Name']] = row.to_dict()
    for idx, row in circle_df.iterrows():
        if not video_name_nesting:
            out[row['Name']] = row.to_dict()
        else:
            if row['Video'] not in list(out.keys()):
                out[row['Video']] = {}
            out[row['Video']][row['Name']] = row.to_dict()
    for idx, row in polygon_df.iterrows():
        if not video_name_nesting:
            out[row['Name']] = row.to_dict()
        else:
            if row['Video'] not in list(out.keys()):
                out[row['Video']] = {}
            out[row['Video']][row['Name']] = row.to_dict()
    return out



def get_roi_data(roi_path: Union[str, os.PathLike], video_name: str) -> tuple:
    """ Helper to read in ROI data from disk"""
    rectangles_df, circles_df, polygon_df = pd.DataFrame(columns=get_rectangle_df_headers()), pd.DataFrame(columns=get_circle_df_headers()), pd.DataFrame(columns=get_polygon_df_headers())
    other_roi_dict = {}
    roi_names, roi_dict = [''], {}
    other_video_names_w_rois = ['']

    if os.path.isfile(roi_path):
        in_rectangles_df, in_circles_df, in_polygon_df = read_roi_data(roi_path=roi_path)
        other_video_names_w_rois = list(set(list(in_rectangles_df['Video'].unique()) + list(in_circles_df['Video'].unique()) + list(in_polygon_df['Video'].unique())))
        other_video_names_w_rois = [x for x in other_video_names_w_rois if x != video_name]
        if len(other_video_names_w_rois) == 0: other_video_names_w_rois = ['']
        rectangles_df = in_rectangles_df[in_rectangles_df['Video'] == video_name].reset_index(drop=True)
        circles_df = in_circles_df[in_circles_df['Video'] == video_name].reset_index(drop=True)
        polygon_df = in_polygon_df[in_polygon_df['Video'] == video_name].reset_index(drop=True)
        other_rectangles_df = in_rectangles_df[in_rectangles_df['Video'] != video_name].reset_index(drop=True)
        other_circles_df = in_circles_df[in_circles_df['Video'] != video_name].reset_index(drop=True)
        other_polygon_df = in_polygon_df[in_polygon_df['Video'] != video_name].reset_index(drop=True)
        other_roi_dict = get_roi_dict_from_dfs(rectangle_df=other_rectangles_df, circle_df=other_circles_df, polygon_df=other_polygon_df, video_name_nesting=True)
        if len(rectangles_df) + len(circles_df) + len(polygon_df) > 0:
            roi_names = list(set(list(rectangles_df['Name'].unique()) + list(circles_df['Name'].unique()) + list(polygon_df['Name'].unique())))
            roi_dict = get_roi_dict_from_dfs(rectangle_df=rectangles_df, circle_df=circles_df, polygon_df=polygon_df, video_name_nesting=False)

    return (rectangles_df, circles_df, polygon_df, roi_dict, roi_names, other_roi_dict, other_video_names_w_rois)


def get_roi_data_for_video_name(roi_path: Union[str, os.PathLike], video_name: str):
    in_rectangles_df, in_circles_df, in_polygon_df = read_roi_data(roi_path=roi_path)
    rectangles_df = in_rectangles_df[in_rectangles_df['Video'] == video_name].reset_index(drop=True)
    circles_df = in_circles_df[in_circles_df['Video'] == video_name].reset_index(drop=True)
    polygon_df = in_polygon_df[in_polygon_df['Video'] == video_name].reset_index(drop=True)
    roi_dict = get_roi_dict_from_dfs(rectangle_df=rectangles_df, circle_df=circles_df, polygon_df=polygon_df, video_name_nesting=False)
    return roi_dict



def get_video_roi_data_from_dict(roi_dict: dict, video_name: str) -> dict:
    out = {}
    for video_key, data in roi_dict.items():
        if video_key == video_name:
            for shape_name, shape_data in data.items():
                out[shape_name] = shape_data
    return out


def change_roi_dict_video_name(roi_dict: dict, video_name: str) -> dict:
    out = {}
    for shape_name, shape_data in roi_dict.items():
        new_shape_data = copy(shape_data)
        new_shape_data['Video'] = video_name
        out[shape_name] = new_shape_data
    return out




def get_ear_tags_for_rectangle(center: Tuple[int, int], width: int, height: int) -> Dict[str, Union[int, Tuple[int, int]]]:
    """
    Knowing the center, width, and height of rectangle, return its vertices.

    :param Tuple[int, int] center: The center x and y coordinates of the rectangle
    :param int width: The width of the rectangle in pixels.
    :param Tuple[int, int] width: The width of the rectangle in pixels.
    :return: Dictionary holding the name of the tag as key and coordinates as value.
    :rtype: Dict[str, Union[int, Tuple[int, int]]]
    """

    check_valid_tuple(x=center, source=get_ear_tags_for_rectangle.__name__, accepted_lengths=(2,), valid_dtypes=Formats.NUMERIC_DTYPES.value)
    check_int(name='width', value=width, min_value=1)
    check_int(name='height', value=height, min_value=1)
    tags = {}
    tags['Top left tag'] = (round((center[1] - (width/2))), round(center[0] - (height/2)))
    tags['Bottom right tag'] = (round(center[1] + (width/2)), round(center[0] + (height/2)))
    tags['Top right tag'] = (round(center[1] + (width/2)), round(center[0] - (height/2)))
    tags['Bottom left tag'] = (round(center[1] - (width / 2)), round(center[0] + (height / 2)))
    tags['Top tag'] = (round(center[1]), round(center[0] - (height / 2)))
    tags['Right tag'] = (round(center[1] + (width / 2)), round(center[0]))
    tags['Left tag'] = (round(center[1] - (width / 2)), round(center[0]))
    tags['Bottom tag'] = (round(center[1]), round(center[0] + (height / 2)))
    return tags



def get_vertices_hexagon(center: Tuple[int, int],
                         radius: int) -> Tuple[np.ndarray, Dict[str, Tuple[int, int]]]:
    """
    Generates the vertices of a regular hexagon centered at a given point with a specified radius.

    :param Tuple[int, int] center:  A tuple (x, y) representing the center coordinates of the hexagon. Must contain exactly two numeric values.
    :param int radius: The radius of the hexagon, which represents the distance from the center to any of the vertices. Must be a positive integer.
    :return: A tuple containing vertices as array and dict.
    :rtype: Tuple[np.ndarray, Dict[str, Tuple[int, int]]]
    """

    check_valid_tuple(x=center, source=get_vertices_hexagon.__name__, accepted_lengths=(2,), valid_dtypes=Formats.NUMERIC_DTYPES.value)
    check_int(name='radius', value=radius, min_value=1)
    vertices = []
    x_c, y_c = center
    for i in range(6):
        angle_rad = np.deg2rad(60 * i)
        x_i = x_c + radius * np.cos(angle_rad)
        y_i = y_c + radius * np.sin(angle_rad)
        vertices.append((x_i, y_i))

    vertices_dict = {"Center_tag": (center[0], center[1])}
    for tag_id, tag in enumerate(vertices):
        vertices_dict[f"Tag_{tag_id}"] = (round(tag[0]), round(tag[1]))
    return (np.round(np.array(vertices)).astype(np.int32), vertices_dict)


def get_half_circle_vertices(center: Tuple[int, int],
                             radius: int,
                             direction: str,
                             n_points: Optional[int] = 50) -> Tuple[np.ndarray, Dict[str, Tuple[int, int]]]:

    """
    Generates vertices for a half-circle with a given radius and direction, centered at a specific point.

    :param Tuple[int, int] center: A tuple (x, y) representing the center coordinates of the half-circle. Must contain exactly two numeric values.
    :param int radius: The radius of the half-circle. Must be a positive integer.
    :param str direction: The direction in which the half-circle is oriented.
    :param Optional[int] n_points: The number of vertices used to approximate the half-circle. Defaults to 50.
    :return: A tuple containing vertices as array and dict.
    :rtype: Tuple[np.ndarray, Dict[str, Tuple[int, int]]]
    """

    check_valid_tuple(x=center, source=get_vertices_hexagon.__name__, accepted_lengths=(2,), valid_dtypes=Formats.NUMERIC_DTYPES.value)
    check_int(name='radius', value=radius, min_value=1)
    check_str(name='direction', options=['NORTH', 'SOUTH', 'WEST', 'EAST', 'NORTH-EAST', 'NORTH-WEST', 'SOUTH-EAST', 'SOUTH-WEST'], value=direction)
    x_c, y_c = center
    if direction == "WEST":
        a = np.linspace(np.pi / 2, 3 * np.pi / 2, n_points)
    elif direction == "EAST":
        a = np.linspace(-np.pi / 2, np.pi / 2, n_points)
    elif direction == "SOUTH":
        a = np.linspace(0, np.pi, n_points)
    elif direction == 'SOUTH-WEST':
        a = np.linspace(np.pi / 4, 5 * np.pi / 4, n_points)
    elif direction == 'SOUTH-EAST':
        a = np.linspace(-np.pi / 4, 3 * np.pi / 4, n_points)
    elif direction == 'NORTH-WEST':
        a = np.linspace(3 * np.pi / 4, 7 * np.pi / 4, n_points)
    elif direction == 'NORTH-EAST':
        a = np.linspace(5 * np.pi / 4, 9 * np.pi / 4, n_points)
    else:
        a = np.linspace(np.pi, 2 * np.pi, n_points)
    x, y = x_c + radius * np.cos(a), y_c + radius * np.sin(a)
    vertices = np.round(np.column_stack((x, y))).astype(np.int32)
    shape_center = np.round(np.array(Polygon(vertices).centroid.coords)[0]).astype(np.int32)
    vertices_dict = {"Center_tag": (shape_center[0], shape_center[1])}

    for tag_id in range(vertices.shape[0]):
        vertices_dict[f"Tag_{tag_id}"] = (vertices[tag_id][0], vertices[tag_id][1])
    return (np.round(np.array(vertices)).astype(np.int32), vertices_dict)



def get_triangle_vertices(center: Tuple[int, int], side_length: int, direction: int) -> Tuple[np.ndarray, Dict[str, Tuple[int, int]]]:
    """
    Find equilateral triangle vertices knowing the center, direction and length side.

    :param Tuple[int, int] center: A tuple (x, y) representing the center coordinates of the hexagon. Must contain exactly two numeric values.
    :param int side_length: The length of each of the three sides in pixels.
    :param int direction: The direction to point the vertices.
    :return: 2-part tuple containing vertices as array and dict.
    :rtype: Tuple[np.ndarray, Dict[str, Tuple[int, int]]]
    """
    check_valid_tuple(x=center, source=get_vertices_hexagon.__name__, accepted_lengths=(2,), valid_dtypes=Formats.NUMERIC_DTYPES.value)
    check_int(name='side_length', value=side_length, min_value=1)
    check_int(name='direction', value=direction, min_value=0, max_value=360)

    direction_radians = np.radians(direction)
    radius = side_length / np.sqrt(3)

    top_vertex = (center[0] + radius * np.cos(direction_radians), center[1] + radius * np.sin(direction_radians))
    vertex2 = (center[0] + radius * np.cos(direction_radians + np.radians(120)), center[1] + radius * np.sin(direction_radians + np.radians(120)))
    vertex3 = (center[0] + radius * np.cos(direction_radians + np.radians(-120)), center[1] + radius * np.sin(direction_radians + np.radians(-120)))
    vertices = np.round(np.array([top_vertex, vertex2, vertex3, top_vertex])).astype(np.int32)
    vertices_dict = {"Center_tag": (int(center[0]), int(center[1])), 'Tag_0':  (int(top_vertex[0]), int(top_vertex[1])), 'Tag_1': (int(vertex2[0]), int(vertex2[1])), 'Tag_2': (int(vertex3[0]), int(vertex3[1]))}
    return (vertices, vertices_dict)



def multiply_ROIs(filename: Union[str, os.PathLike],
                  config_path: Optional[Union[str, os.PathLike]] = None,
                  roi_coordinates_path: Optional[Union[str, os.PathLike]] = None,
                  videos_dir: Optional[Union[str, os.PathLike]] = None) -> None:

    """
    Reproduce ROIs in one video to all other videos in SimBA project.

    :param Union[str, os.PathLike] config_path: Path to SimBA project config file.
    :param Union[str, os.PathLike] filename: Path to video in project for which ROIs should be duplicated in the other videos in the project
    :return: None. The results are stored in the ``/project_folder/logs/measures\ROI_definitions.h5`` of the SimBA project

    :example:
    >>> multiply_ROIs(config_path=r"C:\troubleshooting\mitra\project_folder\project_config.ini", filename=r"C:\troubleshooting\mitra\project_folder\videos\501_MA142_Gi_CNO_0514.mp4")
    """

    if config_path is not None:
        check_file_exist_and_readable(file_path=config_path)
        config = read_config_file(config_path=config_path)
        project_path = config.get(ConfigKey.GENERAL_SETTINGS.value, ConfigKey.PROJECT_PATH.value)
        videos_dir = os.path.join(project_path, "videos")
        roi_coordinates_path = os.path.join(project_path, "logs", Paths.ROI_DEFINITIONS.value)

    check_file_exist_and_readable(file_path=filename)
    _, video_name, video_ext = get_fn_ext(filename)

    if not os.path.isdir(videos_dir):
        raise NotDirectoryError(msg=f'Could not find the videos directory in the SimBA project. SimBA expected a directory at location: {videos_dir}')
    if not os.path.isfile(roi_coordinates_path):
        raise NoROIDataError(msg=f"Cannot multiply ROI definitions: no ROI definitions exist in SimBA project. Could find find a file at expected location {roi_coordinates_path}", source=multiply_ROIs.__name__)

    with pd.HDFStore(roi_coordinates_path) as hdf: roi_data_keys = [x[1:] for x in hdf.keys()]
    missing_keys = [x for x in roi_data_keys if x not in [Keys.ROI_RECTANGLES.value, Keys.ROI_CIRCLES.value, Keys.ROI_POLYGONS.value]]
    if len(missing_keys) > 0:
        raise NoROIDataError(msg=f'The ROI data file {roi_coordinates_path} is corrupted. Missing the following keys: {missing_keys}', source=multiply_ROIs.__name__)

    rectangles_df = pd.read_hdf(path_or_buf=roi_coordinates_path, key=Keys.ROI_RECTANGLES.value)
    circles_df = pd.read_hdf(path_or_buf=roi_coordinates_path, key=Keys.ROI_CIRCLES.value)
    polygon_df = pd.read_hdf(path_or_buf=roi_coordinates_path, key=Keys.ROI_POLYGONS.value)

    check_valid_dataframe(df=rectangles_df, source=f'{multiply_ROIs.__name__} rectangles_df', required_fields=['Video', 'Name'])
    check_valid_dataframe(df=circles_df, source=f'{multiply_ROIs.__name__} circles_df', required_fields=['Video', 'Name'])
    check_valid_dataframe(df=polygon_df, source=f'{multiply_ROIs.__name__} polygon_df', required_fields=['Video', 'Name'])

    videos_w_rectangles = list(rectangles_df["Video"].unique())
    videos_w_circles = list(circles_df["Video"].unique())
    videos_w_polygons = list(polygon_df["Video"].unique())
    videos_w_shapes = list(set(videos_w_rectangles + videos_w_circles + videos_w_polygons))
    if video_name not in videos_w_shapes:
        raise NoROIDataError(msg=f"Cannot replicate ROIs to all other videos: no ROI records exist for {video_name}. Create ROIs for for video {video_name}", source=multiply_ROIs.__name__)

    other_video_file_paths = find_files_of_filetypes_in_directory(directory=videos_dir, extensions=Options.ALL_VIDEO_FORMAT_OPTIONS.value)
    other_video_file_paths = [x for x in other_video_file_paths if x != filename]
    if len(other_video_file_paths) == 0:
        raise NoROIDataError(msg=f"Cannot replicate ROIs to other videos. No other videos exist in project {videos_dir} directory.", source=multiply_ROIs.__name__)

    r_df = [pd.DataFrame(columns=get_rectangle_df_headers()) if video_name not in videos_w_rectangles else rectangles_df[rectangles_df["Video"] == video_name]][0]
    c_df = [pd.DataFrame(columns=get_circle_df_headers()) if video_name not in videos_w_circles else circles_df[circles_df["Video"] == video_name]][0]
    p_df = [pd.DataFrame(columns=get_polygon_df_headers()) if video_name not in videos_w_polygons else polygon_df[polygon_df["Video"] == video_name]][0]

    rectangle_results, circle_results, polygon_results = deepcopy(r_df), deepcopy(c_df), deepcopy(p_df)
    for other_video_file_name in other_video_file_paths:
        _, other_vid_name, ext = get_fn_ext(other_video_file_name)
        if len(r_df) > 0:
            x = deepcopy(r_df); x['Video'] = other_vid_name
            rectangle_results = pd.concat([rectangle_results, x], axis=0)
        if len(circle_results) > 0:
            x = deepcopy(c_df); x['Video'] = other_vid_name
            circle_results = pd.concat([circle_results, x], axis=0)
        if len(polygon_results) > 0:
            x = deepcopy(p_df); x['Video'] = other_vid_name
            polygon_results = pd.concat([polygon_results, x], axis=0)

    rectangle_results = rectangle_results.drop_duplicates(subset=["Video", "Name"], keep="first")
    circle_results = circle_results.drop_duplicates(subset=["Video", "Name"], keep="first")
    polygon_results = polygon_results.drop_duplicates(subset=["Video", "Name"], keep="first")

    store = pd.HDFStore(roi_coordinates_path, mode="w")
    store[Keys.ROI_RECTANGLES.value] = rectangle_results
    store[Keys.ROI_CIRCLES.value] = circle_results
    store[Keys.ROI_POLYGONS.value] = polygon_results
    store.close()
    stdout_success(msg=f"ROIs for {video_name} applied to a further {len(other_video_file_paths)} videos (Duplicated rectangles count: {len(r_df)}, circles: {len(c_df)}, polygons: {len(p_df)}).")
    #print('Next, click on "DRAW" to modify ROI location(s) or click on "RESET" to remove ROI drawing(s)')

def reset_video_ROIs(config_path: Union[str, os.PathLike],
                     filename: Union[str, os.PathLike]) -> None:

    """
    Delete drawn ROIs for a single video in a SimBA project.

    :param Union[str, os.PathLike] config_path: Path to SimBA project config file.
    :param Union[str, os.PathLike] filename: Path to video in project for which ROIs should be duplicated in the other videos in the project
    :return: None. The results are stored in the ``/project_folder/logs/measures\ROI_definitions.h5`` of the SimBA project

    :example:
    >>> reset_video_ROIs(config_path=r"C:\troubleshooting\mitra\project_folder\project_config.ini", filename=r"C:\troubleshooting\mitra\project_folder\videos\501_MA142_Gi_CNO_0514.mp4")
    """

    check_file_exist_and_readable(file_path=config_path)
    check_file_exist_and_readable(file_path=filename)
    _, video_name, video_ext = get_fn_ext(filename)
    config = read_config_file(config_path=config_path)
    project_path = config.get(ConfigKey.GENERAL_SETTINGS.value, ConfigKey.PROJECT_PATH.value)
    roi_coordinates_path = os.path.join(project_path, "logs", Paths.ROI_DEFINITIONS.value)
    if not os.path.isfile(roi_coordinates_path):
        raise NoROIDataError(msg=f"Cannot reset/delete ROI definitions: no ROI definitions exist in SimBA project. Could find find a file at expected location {roi_coordinates_path}. Create ROIs before deleting ROIs.", source=reset_video_ROIs.__name__)
    with pd.HDFStore(roi_coordinates_path) as hdf: roi_data_keys = [x[1:] for x in hdf.keys()]
    missing_keys = [x for x in roi_data_keys if x not in [Keys.ROI_RECTANGLES.value, Keys.ROI_CIRCLES.value, Keys.ROI_POLYGONS.value]]
    if len(missing_keys) > 0:
        raise NoROIDataError(msg=f'The ROI data file {roi_coordinates_path} is corrupted. Missing the following keys: {missing_keys}', source=reset_video_ROIs.__name__)

    rectangles_df = pd.read_hdf(path_or_buf=roi_coordinates_path, key=Keys.ROI_RECTANGLES.value)
    circles_df = pd.read_hdf(path_or_buf=roi_coordinates_path, key=Keys.ROI_CIRCLES.value)
    polygon_df = pd.read_hdf(path_or_buf=roi_coordinates_path, key=Keys.ROI_POLYGONS.value)
    check_valid_dataframe(df=rectangles_df, source=f'{reset_video_ROIs.__name__} rectangles_df', required_fields=['Video'])
    check_valid_dataframe(df=circles_df, source=f'{reset_video_ROIs.__name__} circles_df', required_fields=['Video'])
    check_valid_dataframe(df=polygon_df, source=f'{reset_video_ROIs.__name__} polygon_df', required_fields=['Video'])
    video_rectangle_roi_records = rectangles_df[rectangles_df["Video"] == video_name]
    video_circle_roi_records = circles_df[circles_df["Video"] == video_name]
    video_polygon_roi_records = polygon_df[polygon_df["Video"] == video_name]
    video_roi_cnt = len(video_rectangle_roi_records) + len(video_circle_roi_records) + len(video_polygon_roi_records)
    if video_roi_cnt == 0:
        raise NoROIDataError(msg=f"Cannot delete ROIs for video {video_name}: no ROI records exist for {video_name}. Create ROIs for for video {video_name} first", source=reset_video_ROIs.__name__)

    store = pd.HDFStore(roi_coordinates_path, mode="w")
    store[Keys.ROI_RECTANGLES.value] = rectangles_df[rectangles_df["Video"] != video_name]
    store[Keys.ROI_CIRCLES.value] = circles_df[circles_df["Video"] != video_name]
    store[Keys.ROI_POLYGONS.value] = polygon_df[polygon_df["Video"] != video_name]
    store.close()
    stdout_trash(msg=f"Deleted ROI records for video {video_name}. Deleted rectangle count: {len(video_rectangle_roi_records)}, circles: {len(video_circle_roi_records)}, polygons: {len(video_polygon_roi_records)}.")

def get_image_from_label(tk_lbl: Label):
    """ Given a tkinter label with an image, retrieve image in array format"""

    if not hasattr(tk_lbl, 'image'):
        raise InvalidInputError(msg=f'The label {tk_lbl} does not have a valid image')
    else:
        tk_img = tk_lbl.image
        pil_image = ImageTk.getimage(tk_img)
        img = np.asarray(pil_image)
        return cv2.cvtColor(img, cv2.COLOR_BGR2RGB)


def get_pose_for_roi_ui(pose_path: Union[str, os.PathLike],
                        video_path: Union[str, os.PathLike]) -> Union[None, np.ndarray]:
    video_meta_data = get_video_meta_data(video_path=video_path, raise_error=False)
    if video_meta_data is None:
        VideoFileWarning(msg=f'Cannot plot pose on ROI as cannot read meta data for video {video_path}.', source=get_pose_for_roi_ui.__name__)
        return None
    file_readable = check_file_exist_and_readable(file_path=pose_path, raise_error=False)
    if not file_readable:
        NoFileFoundWarning(msg=f'Cannot plot tracking in ROI window: {pose_path} is unreadable.', source=get_pose_for_roi_ui.__name__)
        return None
    pose_df = read_df(file_path=pose_path, file_type='csv')
    data_align = check_video_and_data_frm_count_align(video=video_path, data=pose_path, raise_error=False)
    if not data_align:
        FrameRangeWarning(msg=f'Cannot plot tracking in ROI window: The data contains {len(pose_path)} frames and the video has {video_meta_data["frame_count"]} frames', source=get_pose_for_roi_ui.__name__)
        return None
    pose_df = pose_df.drop(pose_df.columns[2::3], axis=1)
    pose_data = pose_df.values.reshape(len(pose_df), round(len(pose_df.columns) / 2), 2).astype(np.int32)
    valid_pose = check_valid_array(data=pose_data, source=f'{get_pose_for_roi_ui.__name__} pose_data', accepted_ndims=(3,), accepted_axis_0_shape=[video_meta_data['frame_count']], accepted_dtypes=Formats.INTEGER_DTYPES.value, raise_error=False)
    if not valid_pose:
        FrameRangeWarning(msg=f'Cannot plot tracking in ROI window: The pose data from path {pose_path} is not a 3D numeric array with length {video_meta_data["frame_count"]}', source=get_pose_for_roi_ui.__name__)
    return pose_data


def insert_gridlines_on_roi_img(img: np.ndarray,
                                grid: List[Polygon],
                                color: Tuple[int, int, int],
                                thickness: int) -> np.ndarray:

    if grid is None or len(grid) == 0:
        return img
    else:
        try:
            for polygon in grid:
                cords = np.round(np.array(polygon.exterior.coords)).astype(np.int32)
                img = cv2.polylines(img=img, pts=[cords], isClosed=True, color=color, thickness=thickness, lineType=8)
            return img
        except Exception as e:
            msg = f'Cannot draw gridlines: {e.args}'
            raise InvalidInputError(msg=msg, source=f'{insert_gridlines_on_roi_img.__name__} draw')


__author__ = "Simon Nilsson"

import glob
import os
from collections import defaultdict
from typing import List, Union

import numpy as np
import pandas as pd

try:
    from typing import Literal
except:
    from typing_extensions import Literal

from simba.mixins.config_reader import ConfigReader
from simba.utils.checks import (
    check_all_file_names_are_represented_in_video_log,
    check_file_exist_and_readable, check_if_filepath_list_is_empty, check_int,
    check_that_column_exist, check_valid_boolean, check_valid_lst)
from simba.utils.data import detect_bouts
from simba.utils.enums import Options
from simba.utils.errors import (InvalidInputError, NoChoosenMeasurementError,
                                NoDataError)
from simba.utils.printing import stdout_success
from simba.utils.read_write import get_fn_ext, read_df

FIRST_OCCURRENCE = "First occurrence (s)"
EVENT_COUNT = "Event count"
TOTAL_EVENT_DURATION = "Total event duration (s)"
MEAN_EVENT_DURATION = "Mean event duration (s)"
MEDIAN_EVENT_DURATION = "Median event duration (s)"
MEAN_EVENT_INTERVAL = "Mean event interval (s)"
MEDIAN_EVENT_INTERVAL = "Median event interval (s)"
MEASUREMENT_NAMES = [FIRST_OCCURRENCE, EVENT_COUNT, TOTAL_EVENT_DURATION, MEAN_EVENT_DURATION, MEDIAN_EVENT_DURATION, MEAN_EVENT_INTERVAL, MEDIAN_EVENT_INTERVAL]

class TimeBinsClfCalculator(ConfigReader):
    """
    Computes aggregate classification results in user-defined time-bins. Results are stored in
    the ``project_folder/logs`` directory of the SimBA project`

    :parameter str config_path: path to SimBA project config file in Configparser format.
    :parameter int bin_length: Integer representing the time bin size in seconds
    :parameter List[str] measurements: Aggregate statistic measures calculated for each time bin. OPTIONS: ['First occurrence (s)', 'Event count',
        Total event duration (s)', 'Mean event duration (s)', 'Median event duration (s)', 'Mean event interval (s)',
        'Median event interval (s)']
    :parameter List[str] classifiers: Names of classifiers to calculate aggregate statistics in time-bins for. EXAMPLE: ['Attack', 'Sniffing']

    .. note::
    `Tutorial <https://github.com/sgoldenlab/simba/blob/master/docs/Scenario2.md#part-4--analyze-machine-results>`__.

    Example
    ----------
    >>> timebin_clf_analyzer = TimeBinsClfCalculator(config_path='MyConfigPath', bin_length=15, measurements=['Event count', 'Total event duration (s)'])
    >>> timebin_clf_analyzer.run()

    """

    def __init__(self,
                 config_path: Union[str, os.PathLike],
                 bin_length: int,
                 classifiers: List[str],
                 first_occurrence: bool = True,
                 event_count: bool = True,
                 total_event_duration: bool = True,
                 mean_event_duration: bool = True,
                 median_event_duration: bool = True,
                 mean_interval_duration: bool = True,
                 median_interval_duration: bool = True):

        super().__init__(config_path=config_path)
        check_file_exist_and_readable(file_path=config_path)
        check_int(name=f'{self.__class__.__name__} bin_length', value=bin_length, min_value=1)
        check_valid_lst(data=classifiers, source=f'{self.__class__.__name__} classifiers', valid_dtypes=(str,), valid_values=self.clf_names, min_len=1)
        if len(self.machine_results_paths) == 0:
            raise NoDataError(msg=f'No data files found in {self.machine_results_dir} directory. Get classification data before analyzing classification time-bin data.', source=self.__class__.__name__)
        self.clfs, self.bin_length = classifiers, bin_length
        self.first_occurrence = first_occurrence
        self.event_count = event_count
        self.total_event_duration = total_event_duration
        self.mean_event_duration = mean_event_duration
        self.median_event_duration = median_event_duration
        self.mean_interval_duration = mean_interval_duration
        self.median_interval_duration = median_interval_duration
        self.measurements = []
        for i, j in zip([first_occurrence, event_count, total_event_duration, mean_event_duration, median_event_duration, mean_interval_duration, median_interval_duration], MEASUREMENT_NAMES):
            check_valid_boolean(value=i, source=f'{self.__class__.__name__} {j}', raise_error=True)
            if i: self.measurements.append(j)




    def run(self):
        video_dict = {}

        check_all_file_names_are_represented_in_video_log(video_info_df=self.video_info_df, data_paths=self.machine_results_paths)
        for file_cnt, file_path in enumerate(self.machine_results_paths):
            _, file_name, _ = get_fn_ext(file_path)
            video_dict[file_name] = {}
            print(f'Analyzing classification in time-bins ({self.bin_length}s) for video {file_name} ({file_cnt+1}/{len(self.machine_results_paths)})')
            data_df = read_df(file_path, self.file_type)
            check_that_column_exist(df=data_df, column_name=self.clfs, file_name=file_path)
            video_settings, px_per_mm, fps = self.read_video_info(video_name=file_name)
            bin_frame_length = max(1, int(self.bin_length * fps))
            splits = np.arange(0, data_df.shape[0], bin_frame_length)
            data_df_lst = [data_df.iloc[start: start + bin_frame_length] for start in splits]
            for bin_cnt, df in enumerate(data_df_lst):
                video_dict[file_name][bin_cnt] = {}
                bouts_df = detect_bouts(data_df=df, target_lst=list(self.clf_names), fps=fps)
                bouts_df["Shifted start"] = bouts_df["Start_time"].shift(-1)
                bouts_df["Interval duration"] = (bouts_df["Shifted start"] - bouts_df["End Time"])
                for clf in self.clf_names:
                    video_dict[file_name][bin_cnt][clf] = defaultdict(list)
                    bout_df = bouts_df.loc[bouts_df["Event"] == clf]
                    if len(bouts_df) > 0:
                        video_dict[file_name][bin_cnt][clf][FIRST_OCCURRENCE] = (round(bout_df["Start_time"].min(), 3))
                        video_dict[file_name][bin_cnt][clf][EVENT_COUNT] = len(bout_df)
                        video_dict[file_name][bin_cnt][clf][TOTAL_EVENT_DURATION] = round(bout_df["Bout_time"].sum(), 3)
                        video_dict[file_name][bin_cnt][clf][MEAN_EVENT_DURATION] = round(bout_df["Bout_time"].mean(), 3)
                        video_dict[file_name][bin_cnt][clf][MEDIAN_EVENT_DURATION] = round(bout_df["Bout_time"].median(), 3)
                    else:
                        video_dict[file_name][bin_cnt][clf][FIRST_OCCURRENCE] = None
                        video_dict[file_name][bin_cnt][clf][EVENT_COUNT] = 0
                        video_dict[file_name][bin_cnt][clf][TOTAL_EVENT_DURATION] = 0
                        video_dict[file_name][bin_cnt][clf][MEAN_EVENT_DURATION] = 0
                        video_dict[file_name][bin_cnt][clf][MEDIAN_EVENT_DURATION] = 0
                    if len(bouts_df) > 1:
                        video_dict[file_name][bin_cnt][clf][MEAN_EVENT_INTERVAL] = round(bout_df[:-1]["Interval duration"].mean(), 3)
                        video_dict[file_name][bin_cnt][clf][MEDIAN_EVENT_INTERVAL] = round(bout_df[:-1]["Interval duration"].median(), 3)
                    else:
                        video_dict[file_name][bin_cnt][clf][MEAN_EVENT_INTERVAL] = None
                        video_dict[file_name][bin_cnt][clf][MEDIAN_EVENT_INTERVAL] = None

        print('Saving results...')
        self.out_df_lst = []
        for video_name, video_info in video_dict.items():
            for bin_number, bin_data in video_info.items():
                data_df = (pd.DataFrame.from_dict(bin_data).reset_index().rename(columns={"index": "MEASUREMENT"}))
                data_df = pd.melt(data_df, id_vars=["MEASUREMENT"]).rename(columns={"value": "VALUE", "variable": "CLASSIFIER"})
                data_df.insert(loc=0, column="TIME BIN #", value=bin_number)
                data_df.insert(loc=0, column="VIDEO", value=video_name)
                self.out_df_lst.append(data_df)

        out_df = pd.concat(self.out_df_lst, axis=0).sort_values(by=["VIDEO", "TIME BIN #"]).set_index("VIDEO")
        out_df = out_df[out_df["MEASUREMENT"].isin(self.measurements)]
        out_df = out_df[out_df["CLASSIFIER"].isin(self.clfs)]
        self.save_path = os.path.join(self.project_path, "logs", f"Time_bins_ML_results_{self.datetime}.csv")
        out_df.to_csv(self.save_path)
        self.timer.stop_timer()
        stdout_success(msg=f'Classification time-bins results saved at {self.save_path}', elapsed_time=self.timer.elapsed_time_str)

# test = TimeBinsClfCalculator(config_path=r'D:\troubleshooting\mitra\project_folder\project_config.ini',
#                              classifiers=['lay-on-belly'],
#                              bin_length=60)
#
# test.run()



# test = TimeBinsClf(config_path='/Users/simon/Desktop/envs/troubleshooting/two_black_animals_14bp/project_folder/project_config.ini',
#                    bin_length=2,
#                    measurements=['First occurrence (s)', 'Event count', 'Total event duration (s)', 'Mean event duration (s)'],
#                    classifiers=['Attack', 'Sniffing'])
# test.analyze_timebins_clf()


# test = TimeBinsClf(config_path='/Users/simon/Desktop/troubleshooting/light_analyzer/project_folder/project_config.ini',
#                    bin_length=2,
#                    measurements=['First occurrence (s)', 'Event count', 'Total event duration (s)', 'Mean event duration (s)'])
# test.analyze_timebins_clf()



__author__ = "Simon Nilsson"
import functools
import multiprocessing
import os
import platform
from typing import Dict, List, Optional, Union

import cv2
import numpy as np
from numba import jit

from simba.mixins.config_reader import ConfigReader
from simba.mixins.feature_extraction_mixin import FeatureExtractionMixin
from simba.mixins.plotting_mixin import PlottingMixin
from simba.utils.checks import (
    check_all_file_names_are_represented_in_video_log,
    check_file_exist_and_readable, check_instance, check_int, check_valid_lst)
from simba.utils.errors import (CountError, InvalidInputError,
                                NoSpecifiedOutputError)
from simba.utils.lookups import get_color_dict
from simba.utils.printing import SimbaTimer, stdout_success
from simba.utils.read_write import (concatenate_videos_in_folder,
                                    find_core_cnt, get_fn_ext, read_df)


def distance_plotter_mp(
    frm_cnts: np.array,
    distances: np.ndarray,
    colors: List[str],
    video_setting: bool,
    frame_setting: bool,
    video_name: str,
    video_save_dir: str,
    frame_folder_dir: str,
    style_attr: dict,
    fps: int,
):

    group = int(distances[frm_cnts[0], 0])
    video_writer = None
    if video_setting:
        fourcc = cv2.VideoWriter_fourcc(*"DIVX")
        temp_video_save_path = os.path.join(video_save_dir, f"{group}.avi")
        video_writer = cv2.VideoWriter(
            temp_video_save_path,
            fourcc,
            fps,
            (style_attr["width"], style_attr["height"]),
        )

    for frm_cnt in frm_cnts:
        line_data = distances[:frm_cnt, 1:]
        line_data = np.hsplit(line_data, line_data.shape[1])

        img = PlottingMixin.make_line_plot_plotly(
            data=line_data,
            colors=colors,
            width=style_attr["width"],
            height=style_attr["height"],
            line_width=style_attr["line width"],
            font_size=style_attr["font size"],
            title="Animal distances",
            y_lbl="distance (cm)",
            x_lbl="frame count",
            x_lbl_divisor=fps,
            y_max=style_attr["y_max"],
            line_opacity=style_attr["opacity"],
            save_path=None,
        ).astype(np.uint8)
        if video_setting:
            video_writer.write(img[:, :, :3])
        if frame_setting:
            frm_name = os.path.join(frame_folder_dir, f"{frm_cnt}.png")
            cv2.imwrite(frm_name, np.uint8(img))
        print(
            f"Distance frame created: {frm_cnt} (of {distances.shape[0]}), Video: {video_name}, Processing core: {group}"
        )
    if video_setting:
        video_writer.release()
    return group


class DistancePlotterMultiCore(ConfigReader, PlottingMixin):
    """
     Visualize the distances between pose-estimated body-parts (e.g., two animals) through line
     charts. Results are saved as individual line charts, and/or a video of line charts.
     Uses multiprocessing.

     :param str config_path: path to SimBA project config file in Configparser format
     :param bool frame_setting: If True, creates individual frames.
     :param bool video_setting: If True, creates videos.
     :param bool final_img: If True, creates a single .png representing the entire video.
     :param dict style_attr: Video style attributes (font sizes, line opacity etc.)
     :param List[Union[str, os.PathLike]] data_paths: Files to visualize.
     :param dict line_attr: Representing the body-parts to visualize the distance between and their colors.

    .. note::
       `GitHub tutorial/documentation <https://github.com/sgoldenlab/simba/blob/master/docs/tutorial.md#step-11-visualizations>`__.

    .. image:: _static/img/DistancePlotterMultiCore.png
       :width: 600
       :align: center

    .. image:: _static/img/DistancePlotterMultiCore_1.gif
       :width: 600
       :align: center

    :example:
    >>> style_attr = {'width': 640, 'height': 480, 'line width': 6, 'font size': 8, 'opacity': 0.5}
    >>> line_attr = {0: ['Center_1', 'Center_2', 'Green'], 1: ['Ear_left_2', 'Ear_left_1', 'Red']}
    >>> distance_plotter = DistancePlotterMultiCore(config_path=r'/tests_/project_folder/project_config.ini', frame_setting=False, video_setting=True, final_img=True, style_attr=style_attr, line_attr=line_attr,  files_found=['/test_/project_folder/csv/machine_results/Together_1.csv'], core_cnt=5)
    >>> distance_plotter.run()
    """

    def __init__(
        self,
        config_path: Union[str, os.PathLike],
        data_paths: List[Union[str, os.PathLike]],
        frame_setting: bool,
        video_setting: bool,
        final_img: bool,
        style_attr: Dict[str, int],
        line_attr: List[List[str]],
        core_cnt: Optional[int] = -1,
    ):

        if (not frame_setting) and (not video_setting) and (not final_img):
            raise NoSpecifiedOutputError(
                msg="Please choice to create frames and/or video distance plots",
                source=self.__class__.__name__,
            )
        check_int(
            name=f"{self.__class__.__name__} core_cnt",
            value=core_cnt,
            min_value=-1,
            max_value=find_core_cnt()[0],
        )
        if core_cnt == -1:
            core_cnt = find_core_cnt()[0]
        ConfigReader.__init__(self, config_path=config_path)
        PlottingMixin.__init__(self)
        check_instance(
            source=f"{self.__class__.__name__} line_attr",
            instance=line_attr,
            accepted_types=(list,),
        )
        for cnt, i in enumerate(line_attr):
            check_valid_lst(
                source=f"{self.__class__.__name__} line_attr {cnt}",
                data=i,
                valid_dtypes=(str,),
                exact_len=3,
            )
        check_valid_lst(data=data_paths, valid_dtypes=(str,), min_len=1)
        _ = [check_file_exist_and_readable(i) for i in data_paths]
        (
            self.video_setting,
            self.frame_setting,
            self.data_paths,
            self.style_attr,
            self.line_attr,
            self.final_img,
            self.core_cnt,
        ) = (
            video_setting,
            frame_setting,
            data_paths,
            style_attr,
            line_attr,
            final_img,
            core_cnt,
        )
        if not os.path.exists(self.line_plot_dir):
            os.makedirs(self.line_plot_dir)
        self.color_names = get_color_dict()
        if platform.system() == "Darwin":
            multiprocessing.set_start_method("spawn", force=True)

    @staticmethod
    @jit(nopython=True)
    def __insert_group_idx_column(data: np.array, group: int):
        group_col = np.full((data.shape[0], 1), group)
        return np.hstack((group_col, data))

    def run(self):
        print(f"Processing {len(self.data_paths)} video(s)...")
        check_all_file_names_are_represented_in_video_log(
            video_info_df=self.video_info_df, data_paths=self.data_paths
        )
        for file_cnt, file_path in enumerate(self.data_paths):
            video_timer = SimbaTimer(start=True)
            _, video_name, _ = get_fn_ext(file_path)
            data_df = read_df(file_path, self.file_type)
            try:
                data_df.columns = self.bp_headers
            except ValueError:
                raise CountError(
                    msg=f"SimBA expects {self.bp_headers} columns but found {len(data_df)} columns in {file_path}",
                    source=self.__class__.__name__,
                )
            self.video_info, px_per_mm, fps = self.read_video_info(
                video_name=video_name
            )
            self.save_video_folder = os.path.join(self.line_plot_dir, video_name)
            self.temp_folder = os.path.join(self.line_plot_dir, video_name, "temp")
            self.save_frame_folder_dir = os.path.join(self.line_plot_dir, video_name)
            distances = []
            colors = []
            for cnt, i in enumerate(self.line_attr):
                if i[2] not in list(self.color_names.keys()):
                    raise InvalidInputError(
                        msg=f"{i[2]} is not a valid color. Options: {list(self.color_names.keys())}.",
                        source=self.__class__.__name__,
                    )
                colors.append(i[2])
                bp_1, bp_2 = [f"{i[0]}_x", f"{i[0]}_y"], [f"{i[1]}_x", f"{i[1]}_y"]
                if len(list(set(bp_1) - set(data_df.columns))) > 0:
                    raise InvalidInputError(
                        msg=f"Could not find fields {bp_1} in {file_path}",
                        source=self.__class__.__name__,
                    )
                if len(list(set(bp_2) - set(data_df.columns))) > 0:
                    raise InvalidInputError(
                        msg=f"Could not find fields {bp_2} in {file_path}",
                        source=self.__class__.__name__,
                    )
                distances.append(
                    FeatureExtractionMixin.framewise_euclidean_distance(
                        location_1=data_df[bp_1].values,
                        location_2=data_df[bp_2].values,
                        px_per_mm=px_per_mm,
                        centimeter=True,
                    )
                )
            if self.frame_setting:
                if os.path.exists(self.save_frame_folder_dir):
                    self.remove_a_folder(self.save_frame_folder_dir)
                os.makedirs(self.save_frame_folder_dir)
            if self.video_setting:
                self.video_folder = os.path.join(self.line_plot_dir, video_name)
                if os.path.exists(self.temp_folder):
                    self.remove_a_folder(self.temp_folder)
                os.makedirs(self.temp_folder)
                self.save_video_path = os.path.join(
                    self.line_plot_dir, f"{video_name}.mp4"
                )

            if self.final_img:
                _ = PlottingMixin.make_line_plot(
                    data=distances,
                    colors=colors,
                    width=self.style_attr["width"],
                    height=self.style_attr["height"],
                    line_width=self.style_attr["line width"],
                    font_size=self.style_attr["font size"],
                    title="Animal distances",
                    y_lbl="distance (cm)",
                    x_lbl="time (s)",
                    x_lbl_divisor=fps,
                    y_max=self.style_attr["y_max"],
                    line_opacity=self.style_attr["opacity"],
                    save_path=os.path.join(
                        self.line_plot_dir, f"{video_name}_final_distances.png"
                    ),
                )

            if self.video_setting or self.frame_setting:
                if self.style_attr["y_max"] == -1:
                    self.style_attr["y_max"] = max([np.max(x) for x in distances])
                distances = np.stack(distances, axis=1)
                frm_range = np.arange(0, distances.shape[0])
                frm_range = np.array_split(frm_range, self.core_cnt)

                distances = np.array_split(distances, self.core_cnt)
                distances = [
                    self.__insert_group_idx_column(data=i, group=cnt)
                    for cnt, i in enumerate(distances)
                ]
                distances = np.concatenate(distances, axis=0)
                print(
                    f"Creating distance plots, multiprocessing, follow progress in terminal (chunksize: {self.multiprocess_chunksize}, cores: {self.core_cnt})"
                )
                with multiprocessing.Pool(
                    self.core_cnt, maxtasksperchild=self.maxtasksperchild
                ) as pool:
                    constants = functools.partial(
                        distance_plotter_mp,
                        distances=distances,
                        video_setting=self.video_setting,
                        frame_setting=self.frame_setting,
                        video_name=video_name,
                        video_save_dir=self.temp_folder,
                        frame_folder_dir=self.save_frame_folder_dir,
                        style_attr=self.style_attr,
                        colors=colors,
                        fps=fps,
                    )
                    for cnt, result in enumerate(
                        pool.map(
                            constants, frm_range, chunksize=self.multiprocess_chunksize
                        )
                    ):
                        print(f"Frame batch core {result} complete...")
                        pass
                pool.join()
                pool.terminate()
                if self.video_setting:
                    concatenate_videos_in_folder(
                        in_folder=self.temp_folder,
                        save_path=self.save_video_path,
                        video_format="avi",
                    )
                video_timer.stop_timer()
                stdout_success(
                    msg=f"Distance visualizations created for {video_name} saved at {self.line_plot_dir}",
                    elapsed_time=video_timer.elapsed_time_str,
                )
            self.timer.stop_timer()
            stdout_success(
                msg=f"Distance visualizations complete for {len(self.data_paths)} video(s)",
                elapsed_time=self.timer.elapsed_time_str,
            )


# style_attr = {'width': 640, 'height': 480, 'line width': 6, 'font size': 12, 'y_max': -1, 'opacity': 0.5}
# line_attr = [['Center_1', 'Center_2', 'Green'], ['Ear_left_2', 'Ear_right_2', 'Red']]
# test = DistancePlotterMultiCore(config_path=r'/Users/simon/Desktop/envs/simba/troubleshooting/two_black_animals_14bp/project_folder/project_config.ini',
#                                 frame_setting=True,
#                                 video_setting=True,
#                                 style_attr=style_attr,
#                                 final_img=True,
#                                 data_paths=['/Users/simon/Desktop/envs/simba/troubleshooting/two_black_animals_14bp/project_folder/csv/outlier_corrected_movement_location/Together_1.csv'],
#                                 line_attr=line_attr,
#                                 core_cnt=-1)
# test.run()


# style_attr = {'width': 640, 'height': 480, 'line width': 6, 'font size': 8, 'y_max': 'auto', 'opacity': 0.9}
# line_attr = {0: ['Center_1', 'Center_2', 'Green'], 1: ['Ear_left_2', 'Ear_left_1', 'Red']}
#
# test = DistancePlotterMultiCore(config_path=r'/Users/simon/Desktop/envs/troubleshooting/two_black_animals_14bp/project_folder/project_config.ini',
#                                 frame_setting=False,
#                                 video_setting=True,
#                                 style_attr=style_attr,
#                                 final_img=True,
#                                 files_found=['/Users/simon/Desktop/envs/troubleshooting/two_black_animals_14bp/project_folder/csv/machine_results/Together_1.csv'],
#                                 line_attr=line_attr,
#                                 core_cnt=3)
# test.create_distance_plot()
# #
# style_attr = {'width': 640, 'height': 480, 'line width': 6, 'font size': 8}
# line_attr = {0: ['Termite_1_Head_1', 'Termite_1_Thorax_1', 'Dark-red']}
#


# style_attr = {'width': 640, 'height': 480, 'line width': 6, 'font size': 8, 'opacity': 0.5, 'y_max': 'auto'}
# line_attr = {0: ['Center_1', 'Center_2', 'Green'], 1: ['Ear_left_2', 'Ear_left_1', 'Red']}
#
# test = DistancePlotterMultiCore(config_path=r'/Users/simon/Desktop/envs/troubleshooting/two_black_animals_14bp/project_folder/project_config.ini',
#                        frame_setting=False,
#                        video_setting=True,
#                        style_attr=style_attr,
#                                  final_img=False,
#                        files_found=['/Users/simon/Desktop/envs/troubleshooting/two_black_animals_14bp/project_folder/csv/machine_results/Together_1.csv'],
#                        line_attr=line_attr,
#                                 core_cnt=5)
# test.create_distance_plot()

# style_attr = {'width': 640, 'height': 480, 'line width': 6, 'font size': 8}
# line_attr = {0: ['Termite_1_Head_1', 'Termite_1_Thorax_1', 'Dark-red']}

# test = DistancePlotterSingleCore(config_path=r'/Users/simon/Desktop/envs/troubleshooting/Termites_5/project_folder/project_config.ini',
#                                  frame_setting=False,
#                        video_setting=True,
#                        style_attr=style_attr,
#                        files_found=['/Users/simon/Desktop/envs/troubleshooting/Termites_5/project_folder/csv/outlier_corrected_movement_location/termites_1.csv'],
#                        line_attr=line_attr)
# test.create_distance_plot()
